"""
Training an ElasticNet model
=============================================================

This example shows you how to train a simple elasticnet model and use it to submit to the challenge
"""

#############################################################################
# Loading up the data for the competition
# ---------------------------------------
from biolearn.data_library import GeoData

#Download the data file for the warmup challenge linked here https://www.synapse.org/#!Synapse:syn52966292/wiki/625231
DOWNLOADED_DATA_FILE_PATH="ADD YOUR PATH HERE"
challenge_data = GeoData.from_methylation_matrix(DOWNLOADED_DATA_FILE_PATH)
challenge_data


#############################################################################
# Load up some training data
# ------------------------------------------------------
from biolearn.data_library import DataLibrary
data = DataLibrary().get("GSE40279").load()
data.metadata


############################################################################################################################
# Narrow down what sites are correlated with age
# --------------------------------------------------------------------------------------------------------------------------
#NOTE: This takes a long time to run
import numpy as np
from sklearn.linear_model import LinearRegression

# Extract data from your 'data' object
X = data.dnam.transpose().values  # Transpose to have samples as rows and cpg sites as columns
y = data.metadata['age'].values

# Parameters for bootstrap and feature selection
n_bootstrap = 20
threshold = 0.05 

# Store count of times each CpG site is deemed significant
cpg_counts = np.zeros(X.shape[1])

# Begin bootstrap iterations
for _ in range(n_bootstrap):
    # Sample with replacement from X, y
    sample_idx = np.random.choice(range(X.shape[0]), size=X.shape[0], replace=True)
    X_sample = X[sample_idx]
    y_sample = y[sample_idx]
    
    # Train model
    model = LinearRegression()
    model.fit(X_sample, y_sample)
    
    # Identify significant CpG sites (based on magnitude of coefficients)
    significant_cpgs = np.where(np.abs(model.coef_) > threshold)[0]
    cpg_counts[significant_cpgs] += 1

# Determine stable CpG sites
stable_cpg_sites = np.where(cpg_counts > n_bootstrap * 0.6)[0]
stable_cpg_names = data.dnam.index[stable_cpg_sites].tolist()

print(f"Stable CpG sites (associated with age in more than 60% of bootstrap samples): {stable_cpg_sites}")

############################################################################################################################
# Seperate data into training and test sets
# --------------------------------------------------------------------------------------------------------------------------
from sklearn.model_selection import train_test_split

df = data.dnam.transpose()
df['age'] = data.metadata['age']
top_sites_df = df[stable_cpg_names]

X = top_sites_df
y = df['age']

X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=42)

############################################################################################################################
# Train a model using elastic net
# --------------------------------------------------------------------------------------------------------------------------
# Define the model
# alpha is the regularization strength, and l1_ratio defines the mix between L1 and L2
# l1_ratio = 1 is Lasso; l1_ratio = 0 is Ridge.
from sklearn.linear_model import ElasticNet
from sklearn.metrics import mean_squared_error

model = ElasticNet(alpha=0.01, l1_ratio=0.3, max_iter=10000)
# Train the model
model.fit(X_train, y_train)

# Predict and evaluate
y_pred = model.predict(X_test)
mse = mean_squared_error(y_test, y_pred)

print(f"Mean Squared Error on Test Data: {mse}")

############################################################################################################################
# Plot the results to see how good our model is
# --------------------------------------------------------------------------------------------------------------------------
import matplotlib.pyplot as plt

y_pred = model.predict(X_test)

plt.figure(figsize=(10, 8))
plt.scatter(y_test, y_pred, alpha=0.7, edgecolors='w', linewidth=0.5)
plt.plot([min(y_test), max(y_test)], [min(y_test), max(y_test)], 'k--', lw=3)  # y=x line for reference
plt.xlabel('Actual Age')
plt.ylabel('Predicted Age')
plt.title('Actual Age vs. Predicted Age')
plt.grid(True, which='both', linestyle='--', linewidth=0.5)
plt.show()

# Calculate Mean Squared Error (MSE)
mse = np.mean((y_test - y_pred) ** 2)
print(f"Mean Squared Error (MSE): {mse:.4f}")

# Calculate Mean Absolute Error (MAE)
mae = np.mean(np.abs(y_test - y_pred))
print(f"Mean Absolute Error (MAE): {mae:.4f}")

############################################################################################################################
# Run the challenge data through the model
# --------------------------------------------------------------------------------------------------------------------------

pruned_data = challenge_data.dnam.T[stable_cpg_names]
pruned_data = pruned_data.fillna(0)
challenge_results = model.predict(pruned_data)

############################################################################################################################
# Save the results as an output file for submission
# --------------------------------------------------------------------------------------------------------------------------
import pandas as pd

predicted_age_df = pd.DataFrame({
    'predictedAge': challenge_results
}, index=challenge_data.dnam.columns)
predicted_age_df.index.name = 'sampleId'
predicted_age_df


import pandas as pd
import numpy as np
from biolearn.data_library import GeoData


def test_metadata_population():
    data = {"Sample1": [0.1, 0.2, 0.3], "Sample2": [0.4, 0.5, 0.6]}
    dnam_df = pd.DataFrame(data, index=["Site1", "Site2", "Site3"])

    geo_data = GeoData.from_methylation_matrix(dnam_df)

    assert list(geo_data.metadata.index) == list(
        dnam_df.columns
    ), "Metadata row names do not match expected values."


def test_technical_duplicate_averaging():
    data = {"Sample1": [0.1, 0.2], "Sample2": [0.4, 0.5]}
    dnam_df = pd.DataFrame(data, index=["Site1_1", "Site1_2"])

    expected_data = {"Sample1": [0.15], "Sample2": [0.45]}
    expected_df = pd.DataFrame(expected_data, index=["Site1"])

    geo_data = GeoData.from_methylation_matrix(dnam_df)

    pd.testing.assert_frame_equal(
        geo_data.dnam, expected_df, check_dtype=False
    )


def test_can_load_from_file_path(tmp_path):
    # Create a temporary CSV file to simulate file input
    d = tmp_path / "sub"
    d.mkdir()
    file_path = d / "methylation_matrix.csv"
    data = {"Sample1": [0.1, 0.2, 0.3], "Sample2": [0.4, 0.5, 0.6]}
    dnam_df = pd.DataFrame(data, index=["Site1", "Site2", "Site3"])
    dnam_df.to_csv(file_path)

    # Create GeoData instance from file path
    geo_data = GeoData.from_methylation_matrix(str(file_path))

    # Verify the data is loaded correctly
    pd.testing.assert_frame_equal(geo_data.dnam, dnam_df, check_dtype=False)



import pytest
import pandas as pd
import numpy as np

from biolearn.imputation import (
    impute_from_standard,
    impute_from_average,
    hybrid_impute,
)

df_test = pd.DataFrame(
    {
        "Sample1": [1, np.nan, 3, 4],
        "Sample2": [np.nan, 1, 2, 3],
        "Sample3": [1, 2, np.nan, 4],
        "Sample4": [2, 1, 3, 4],
        "Sample5": [2, np.nan, 2, 3],
        "Sample6": [4, 2, 2, 4],
    },
    index=["cpg1", "cpg2", "cpg3", "cpg4"],
)

cpg_averages_test = pd.Series(
    {"cpg1": 1.5, "cpg2": 2.5, "cpg3": 3.5, "cpg4": 4.5, "cpg5": 5.5}
)

required_cpgs = ["cpg1", "cpg2", "cpg3", "cpg4", "cpg5"]


def no_missing_values(df):
    return df.isna().sum().sum() == 0


# Tests
def test_impute_from_standard():
    df_filled = impute_from_standard(df_test, cpg_averages_test)
    assert no_missing_values(df_filled)
    assert df_filled.loc["cpg2", "Sample1"] == 2.5


def test_impute_from_standard_specific_cpgs():
    specific_cpgs = ["cpg1", "cpg3"]
    df_filled = impute_from_standard(
        df_test, cpg_averages_test, cpgs_to_impute=specific_cpgs
    )

    print(df_filled)
    print(df_test)
    assert not df_filled.loc[specific_cpgs].isna().any().any()
    assert (
        df_filled.drop(specific_cpgs).isna().sum().sum()
        == df_test.drop(specific_cpgs).isna().sum().sum()
    )
    assert df_filled.loc["cpg1", "Sample2"] == 1.5
    assert df_filled.loc["cpg3", "Sample3"] == 3.5


def test_impute_from_average():
    df_filled = impute_from_average(df_test)
    assert no_missing_values(df_filled)
    assert (
        df_filled.loc["cpg1", "Sample2"] == 2
    )  # Ensure the value was filled with the row average.


def test_impute_from_average_specific_cpgs():
    specific_cpgs = ["cpg1", "cpg3"]
    df_filled = impute_from_average(df_test, cpgs_to_impute=specific_cpgs)

    print(df_filled)
    print(df_test)
    assert not df_filled.loc[specific_cpgs].isna().any().any()
    assert (
        df_filled.drop(specific_cpgs).isna().sum().sum()
        == df_test.drop(specific_cpgs).isna().sum().sum()
    )
    assert df_filled.loc["cpg1", "Sample2"] == 2
    assert df_filled.loc["cpg3", "Sample3"] == 2.4


def test_impute_from_average_missing_column():
    existing_cpg = "cpg1"
    non_existent_cpg = "cpg5"
    cpgs_to_impute = [existing_cpg, non_existent_cpg]

    df_filled = impute_from_average(df_test, cpgs_to_impute=cpgs_to_impute)

    assert non_existent_cpg not in df_filled.columns
    assert not df_filled.loc[existing_cpg].isna().any()


def test_replacement_with_mean():
    result = hybrid_impute(
        df_test, cpg_averages_test, ["cpg1", "cpg2", "cpg3", "cpg4", "cpg5"]
    )
    assert no_missing_values(result)
    assert result.loc["cpg1", "Sample1"] == 1
    assert result.loc["cpg1", "Sample2"] == 2


def test_replacement_with_source_below_threshold():
    result = hybrid_impute(
        df_test, cpg_averages_test, ["cpg1", "cpg2", "cpg3", "cpg4", "cpg5"]
    )
    assert no_missing_values(result)
    assert result.loc["cpg2", "Sample1"] == 2.5  # Missing value is replaced
    assert result.loc["cpg2", "Sample2"] == 2.5  # Existing value is replaced


def test_addition_of_missing_values():
    result = hybrid_impute(
        df_test, cpg_averages_test, ["cpg1", "cpg2", "cpg3", "cpg4", "cpg5"]
    )
    assert no_missing_values(result)
    assert "cpg5" in result.index
    assert result.loc["cpg5", "Sample1"] == 5.5
    assert result.loc["cpg5", "Sample2"] == 5.5


def test_error_when_required_cpg_not_available():
    with pytest.raises(ValueError):
        hybrid_impute(
            df_test,
            cpg_averages_test.drop("cpg5"),
            ["cpg1", "cpg2", "cpg3", "cpg4", "cpg5"],
        )


def test_modification_of_threshold():
    result = hybrid_impute(
        df_test, cpg_averages_test, ["cpg1", "cpg2", "cpg3", "cpg4", "cpg5"], 1
    )
    # cpg2 for Sample1 should be replaced with source value because threshold is 1
    assert result.loc["cpg2", "Sample1"] == 2.5
    result = hybrid_impute(
        df_test,
        cpg_averages_test,
        ["cpg1", "cpg2", "cpg3", "cpg4", "cpg5"],
        0.5,
    )
    # cpg2 for Sample1 should be replaced with mean value because threshold is 0.5
    assert (
        result.loc["cpg2", "Sample1"] == 1.5
    )  # mean of available value and source value for cpg2


# Run the tests
if __name__ == "__main__":
    pytest.main([__file__])



from biolearn.data_library import DataSource
import pandas as pd
import os
import sys
import hashlib

source_url = "https://ftp.ncbi.nlm.nih.gov/geo/series/GSE41nnn/GSE41169/matrix/GSE41169_series_matrix.txt.gz"

# Paths for the script, data, and metadata files
script_dir = os.path.dirname(__file__)
data_folder_path = os.path.join(script_dir, "data/external/")
data_file_path = os.path.join(data_folder_path, "DNAmTestSet.csv")
metadata_file_path = os.path.join(data_folder_path, "testset_metadata.csv")


expected_hash = (
    "3ee979ae6d2f1572f72f02e951073af0910b5b1576d46177cbfb76932fda6693"
)


def compute_file_hash(file_path, chunk_size=1024 * 1024):
    hash_algorithm = hashlib.sha256()
    with open(file_path, "rb") as f:
        while True:
            data = f.read(chunk_size)
            if not data:
                break
            hash_algorithm.update(data)
    return hash_algorithm.hexdigest()


def is_file_valid(file_path, expected_hash):
    if os.path.exists(file_path):
        actual_hash = compute_file_hash(file_path)
        if actual_hash != expected_hash:
            print(
                f"Warning: Actual file hash {actual_hash} does not equal expected hash {expected_hash}"
            )
        return actual_hash == expected_hash
    return False


def ensure_folder_exists(path):
    # Get the directory part of the path (if it is a file)
    dir_path = os.path.dirname(path)

    if os.path.isdir(path):
        dir_path = path

    if not os.path.exists(dir_path):
        os.makedirs(dir_path)


ensure_folder_exists(data_file_path)

print("Checking if file generation is needed")
if not is_file_valid(data_file_path, expected_hash):
    print("Loading data source")
    data_source_spec = {
        "id": "TestData",
        "path": source_url,
        "parser": {
            "type": "geo-matrix",
            "id-row": 33,
            "metadata": {
                "age": {"row": 47, "parse": "numeric"},
                "sex": {"row": 41, "parse": "sex"},
            },
            "matrix-start": 73,
        },
    }
    source = DataSource(data_source_spec)
    loaded_data = source.load()

    print("Processing DNAm data")
    dnam = (
        loaded_data.dnam.transpose().head(10).transpose()
    )  # Transpose and limit to 10 items
    dnam.to_csv(data_file_path)

    print("Verifying new DNAm data file")
    if not is_file_valid(data_file_path, expected_hash):
        print(
            "ERROR: Generated DNAm data file does not match expected. Exiting."
        )
        sys.exit(1)
    else:
        print("DNAm data file matches existing hash. Done.")

    print("Processing metadata")
    metadata = loaded_data.metadata.head(10)  # Limit to 10 items
    metadata.to_csv(metadata_file_path)
    print("Metadata file generated and saved.")
else:
    print("File matches existing hash. Done.")



import pandas as pd
from biolearn.util import get_data_file


def impute_from_standard(dnam, cpg_averages, cpgs_to_impute=None):
    """
    Impute all missing values in a DNA methylation dataset using the averages from an external dataset.

    Args:
        dnam (pd.DataFrame): DataFrame with samples as columns and CpG sites as rows.
        cpg_averages (pd.Series): Series containing reference averages for CpG sites.
        cpgs_to_impute (list of str, optional): List of CpG sites to impute. Missing cpgs will only be imputed if in this list.

    Returns:
        pd.DataFrame: DataFrame with missing values filled.
    """
    if cpgs_to_impute:
        impute_rows = dnam.loc[cpgs_to_impute]
        impute_rows = impute_rows.apply(lambda col: col.fillna(cpg_averages))
        df_filled = dnam.combine_first(impute_rows)
    else:
        df_filled = dnam.apply(lambda col: col.fillna(cpg_averages))
    return df_filled


def impute_from_average(dnam, cpgs_to_impute=None):
    """
    Impute all missing values in a DNA methylation dataset using the average from the dataset itself.

    Args:
        dnam (pd.DataFrame): DataFrame with samples as columns and CpG sites as rows.
        cpgs_to_impute (list of str, optional): List of CpG sites to impute.

    Returns:
        pd.DataFrame: DataFrame with missing values filled.
    """
    # Protect input from mutation
    dnam_copy = dnam.copy()
    means = dnam_copy.mean(axis=1)

    if cpgs_to_impute:
        # Filter out non-existent CpG sites
        existing_cpgs = [
            cpg for cpg in cpgs_to_impute if cpg in dnam_copy.index
        ]

        # Apply imputation only to existing CpG sites
        mask = dnam_copy.loc[existing_cpgs].isna()
        dnam_copy.loc[existing_cpgs] = dnam_copy.loc[existing_cpgs].where(
            ~mask, means[existing_cpgs], axis=0
        )
    else:
        dnam_copy = dnam_copy.where(dnam_copy.notna(), means, axis=0)

    return dnam_copy


def hybrid_impute(dnam, cpg_source, required_cpgs, threshold=0.8):
    """
    Imputes missing values in a DNA methylation dataset based on a threshold. Sites with data below the threshold are replaced from an external source, while others are imputed using the average of existing values.

    Args:
        dnam (pd.DataFrame): DataFrame with samples as columns and CpG sites as rows.
        cpg_source (pd.Series): Series containing reference values for CpG sites.
        required_cpgs (list of str): List of CpG sites that need to be in the final dataset.
        threshold (float, optional): Threshold for determining imputation strategy. Default is 0.8.

    Returns:
        pd.DataFrame: DataFrame with missing values filled.

    Raises:
        ValueError: If certain required CpG sites are missing from both the dataset and the cpg_source.
    """
    # Drop rows below the threshold, these will be replaced entirely from cpg_source
    cpgs_below_threshold = dnam.notna().mean(axis=1) < threshold
    dnam = dnam.drop(dnam[cpgs_below_threshold].index)

    # Impute remaining rows using impute_from_average
    df_filled = impute_from_average(dnam)

    missing_cpgs_from_dataset = set(required_cpgs) - set(df_filled.index)
    missing_cpgs_from_source = [
        cpg for cpg in missing_cpgs_from_dataset if cpg not in cpg_source
    ]

    if missing_cpgs_from_source:
        raise ValueError(
            f"Tried to fill the following cpgs but they were missing from cpg_source: {missing_cpgs_from_source}"
        )

    for cpg in missing_cpgs_from_dataset:
        df_filled.loc[cpg] = cpg_source.loc[cpg]

    return df_filled.sort_index()



import os
import hashlib
import requests
from urllib.parse import urlparse
import shutil
import appdirs
import pandas as pd


def get_data_file(relative_path):
    script_dir = os.path.dirname(
        __file__
    )  # get the directory of the current script
    data_file_path = os.path.join(
        script_dir, "data", relative_path
    )  # build the path to the data file
    return data_file_path


def get_test_data_file(relative_path):
    script_dir = os.path.dirname(
        __file__
    )  # get the directory of the current script
    data_file_path = os.path.join(
        script_dir, "test/data", relative_path
    )  # build the path to the data file
    return data_file_path


def load_test_data_file(relative_path):
    test_sample = pd.read_csv(get_test_data_file(relative_path), index_col=0)
    return test_sample


def is_url(s):
    """Check if the input string is a URL."""
    parsed = urlparse(s)
    return bool(parsed.scheme)


def cached_download(url_or_filepath):
    """Downloads the file at a URL and saves it locally. If called again with the same URL, it will use the saved file.
    If the input is a local file path and the file does not exist, raises a FileNotFoundError.
    Returns the local filepath."""
    if is_url(url_or_filepath):
        url = url_or_filepath
        url_path = urlparse(url).path
        ext = os.path.splitext(url_path)[1]
        filename = hashlib.sha256(url.encode()).hexdigest() + ext

        app_name = "bio-learn"
        download_path = appdirs.user_cache_dir(app_name)
        os.makedirs(download_path, exist_ok=True)
        filepath = os.path.join(download_path, filename)

        if not os.path.exists(filepath):
            # Download the file if it doesn't already exist locally
            try:
                response = requests.get(url, stream=True)
                response.raise_for_status()  # Check for HTTP errors
                with open(filepath, "wb") as out_file:
                    shutil.copyfileobj(response.raw, out_file)
            except requests.RequestException as e:
                raise Exception(
                    f"Failed to download the file from {url}. Error: {e}"
                )
    else:
        if not os.path.exists(url_or_filepath):
            raise FileNotFoundError(
                f"The file does not exist: {url_or_filepath}"
            )
        filepath = url_or_filepath

    return filepath



import inspect
import os
import subprocess
import sys
from functools import partial
from operator import attrgetter

REVISION_CMD = "git rev-parse --short HEAD"


def _get_git_revision():
    try:
        revision = subprocess.check_output(REVISION_CMD.split()).strip()
    except (subprocess.CalledProcessError, OSError):
        print("Failed to execute git to get revision")
        return None
    return revision.decode("utf-8")


def _linkcode_resolve(domain, info, package, url_fmt, revision):
    # Continue as normal for non-attributes
    if domain not in ("py", "pyx") or not info.get("module") or '.' not in info["fullname"]:
        return None
    
    class_name, _, attr = info["fullname"].partition('.')
    module = __import__(info["module"], fromlist=[class_name])
    
    # If we're trying to get an attribute of a class, skip it
    if attr:
        return None
    
    obj = getattr(module, class_name, None)
    if obj is None:
        return None
    
    obj = inspect.unwrap(obj)

    try:
        fn = inspect.getsourcefile(obj)
    except Exception:
        fn = None
    if not fn:
        try:
            fn = inspect.getsourcefile(sys.modules[obj.__module__])
        except Exception:
            fn = None
    if not fn:
        return

    # Don't include filenames from outside this package's tree
    if os.path.dirname(__import__(package).__file__) not in fn:
        return

    fn = os.path.relpath(
        fn, start=os.path.dirname(__import__(package).__file__)
    )
    try:
        lineno = inspect.getsourcelines(obj)[1]
    except Exception:
        lineno = ""
    return url_fmt.format(
        revision=revision, package=package, path=fn, lineno=lineno
    )


def make_linkcode_resolve(package, url_fmt):
    """Return a linkcode_resolve function for the given URL format.

    revision is a git commit reference (hash or name)

    package is the name of the root module of the package

    url_fmt is along the lines of ('https://github.com/USER/PROJECT/'
                                   'blob/{revision}/{package}/'
                                   '{path}#L{lineno}')
    """
    revision = _get_git_revision()
    return partial(
        _linkcode_resolve, revision=revision, package=package, url_fmt=url_fmt
    )



import numpy as np
import pandas as pd
import cvxpy as cp

from sklearn.linear_model import LinearRegression

from biolearn.data_library import GeoData
from biolearn.dunedin_pace import dunedin_pace_normalization
from biolearn.util import get_data_file


def anti_trafo(x, adult_age=20):
    y = np.where(
        x < 0, (1 + adult_age) * np.exp(x) - 1, (1 + adult_age) * x + adult_age
    )
    return y


model_definitions = {
    "Horvathv1": {
        "year": 2013,
        "species": "Human",
        "tissue": "Multi-tissue",
        "source": "https://genomebiology.biomedcentral.com/articles/10.1186/gb-2013-14-10-r115",
        "output": "Age (Years)",
        "model": {
            "type": "LinearMethylationModel",
            "file": "Horvath1.csv",
            "transform": lambda sum: anti_trafo(sum + 0.696),
        },
    },
    "Hannum": {
        "year": 2013,
        "species": "Human",
        "tissue": "Blood",
        "source": "https://www.sciencedirect.com/science/article/pii/S1097276512008933",
        "output": "Age (Years)",
        "model": {"type": "LinearMethylationModel", "file": "Hannum.csv"},
    },
    "Lin": {
        "year": 2016,
        "species": "Human",
        "tissue": "Blood",
        "source": "https://www.aging-us.com/article/100908/text",
        "output": "Age (Years)",
        "model": {
            "type": "LinearMethylationModel",
            "file": "Lin.csv",
            "transform": lambda sum: sum + 12.2169841,
        },
    },
    "PhenoAge": {
        "year": 2018,
        "species": "Human",
        "tissue": "Blood",
        "source": "https://www.aging-us.com/article/101414/text",
        "output": "Age (Years)",
        "model": {
            "type": "LinearMethylationModel",
            "file": "PhenoAge.csv",
            "transform": lambda sum: sum + 60.664,
        },
    },
    "YingCausAge": {
        "year": 2022,
        "species": "Human",
        "tissue": "Blood",
        "source": "https://www.biorxiv.org/content/10.1101/2022.10.07.511382v2",
        "output": "Age (Years)",
        "model": {
            "type": "LinearMethylationModel",
            "file": "YingCausAge.csv",
            "transform": lambda sum: sum + 86.80816381,
        },
    },
    "YingDamAge": {
        "year": 2022,
        "species": "Human",
        "tissue": "Blood",
        "source": "https://www.biorxiv.org/content/10.1101/2022.10.07.511382v2",
        "output": "Age (Years)",
        "model": {
            "type": "LinearMethylationModel",
            "file": "YingDamAge.csv",
            "transform": lambda sum: sum + 543.4315887,
        },
    },
    "YingAdaptAge": {
        "year": 2022,
        "species": "Human",
        "tissue": "Blood",
        "source": "https://www.biorxiv.org/content/10.1101/2022.10.07.511382v2",
        "output": "Age (Years)",
        "model": {
            "type": "LinearMethylationModel",
            "file": "YingAdaptAge.csv",
            "transform": lambda sum: sum - 511.9742762,
        },
    },
    "Horvathv2": {
        "year": 2018,
        "species": "Human",
        "tissue": "Skin + blood",
        "source": "https://www.aging-us.com/article/101508/text",
        "output": "Age (Years)",
        "model": {
            "type": "LinearMethylationModel",
            "file": "Horvath2.csv",
            "transform": lambda sum: anti_trafo(sum - 0.447119319),
        },
    },
    "PEDBE": {
        "year": 2019,
        "species": "Human",
        "tissue": "Buccal",
        "source": "https://www.pnas.org/doi/10.1073/pnas.1820843116",
        "output": "Age (Years)",
        "model": {
            "type": "LinearMethylationModel",
            "file": "PEDBE.csv",
            "transform": lambda sum: anti_trafo(sum - 2.1),
        },
    },
    "Zhang_10": {
        "year": 2019,
        "species": "Human",
        "tissue": "Blood",
        "source": "https://www.nature.com/articles/ncomms14617",
        "output": "Mortality Risk",
        "model": {"type": "LinearMethylationModel", "file": "Zhang_10.csv"},
    },
    "DunedinPoAm38": {
        "year": 2020,
        "species": "Human",
        "tissue": "Blood",
        "source": "https://elifesciences.org/articles/54870#s2",
        "output": "Aging Rate (Years/Year)",
        "model": {
            "type": "LinearMethylationModel",
            "file": "DunedinPoAm38.csv",
            "transform": lambda sum: sum - 0.06929805,
        },
    },
    "DunedinPACE": {
        "year": 2022,
        "species": "Human",
        "tissue": "Blood",
        "source": "https://www.proquest.com/docview/2634411178",
        "output": "Aging Rate (Years/Year)",
        "model": {
            "type": "LinearMethylationModel",
            "file": "DunedinPACE.csv",
            "transform": lambda sum: sum - 1.949859,
            "preprocess": dunedin_pace_normalization,
            "default_imputation": "none",
        },
    },
    "GrimAgeV1": {
        "year": 2019,
        "species": "Human",
        "tissue": "Blood",
        "source": "https://www.ncbi.nlm.nih.gov/pmc/articles/PMC6366976/",
        "output": "Mortality Adjusted Age (Years)",
        "model": {"type": "GrimageModel", "file": "GrimAgeV1.csv"},
    },
    "GrimAgeV2": {
        "year": 2022,
        "species": "Human",
        "tissue": "Blood",
        "source": "https://www.ncbi.nlm.nih.gov/pmc/articles/PMC9792204/",
        "output": "Mortality Adjusted Age (Years)",
        "model": {"type": "GrimageModel", "file": "GrimAgeV2.csv"},
    },
    "AlcoholMcCartney": {
        "year": 2018,
        "species": "Human",
        "tissue": "Blood",
        "source": "https://www.ncbi.nlm.nih.gov/pmc/articles/PMC6158884/",
        "output": "Alcohol Consumption",
        "model": {"type": "LinearMethylationModel", "file": "Alcohol.csv"},
    },
    "BMI_McCartney": {
        "year": 2018,
        "species": "Human",
        "tissue": "Blood",
        "source": "https://www.ncbi.nlm.nih.gov/pmc/articles/PMC6158884/",
        "output": "BMI",
        "model": {"type": "LinearMethylationModel", "file": "BMI.csv"},
    },
    "DNAmTL": {
        "year": 2019,
        "species": "Human",
        "tissue": "Blood, Adipose",
        "source": "https://www.ncbi.nlm.nih.gov/pmc/articles/PMC6738410/",
        "output": "Telomere Length",
        "model": {
            "type": "LinearMethylationModel",
            "file": "DNAmTL.csv",
            "transform": lambda sum: sum - 7.924780053,
        },
    },
    "HRSInCHPhenoAge": {
        "year": 2022,
        "species": "Human",
        "tissue": "Blood",
        "output": "Age (Years)",
        "source": "https://www.nature.com/articles/s43587-022-00248-2",
        "model": {
            "type": "LinearMethylationModel",
            "file": "HRSInCHPhenoAge.csv",
            "transform": lambda sum: sum + 52.8334080,
        },
    },
    "Knight": {
        "year": 2016,
        "species": "Human",
        "tissue": "Cord Blood",
        "source": "https://genomebiology.biomedcentral.com/articles/10.1186/s13059-016-1068-z",
        "output": "Gestational Age",
        "model": {
            "type": "LinearMethylationModel",
            "file": "Knight.csv",
            "transform": lambda sum: sum + 41.7,
        },
    },
    "LeeControl": {
        "year": 2019,
        "species": "Human",
        "tissue": "Placenta",
        "source": "https://www.aging-us.com/article/102049/text",
        "output": "Gestational Age",
        "model": {
            "type": "LinearMethylationModel",
            "file": "LeeControl.csv",
            "transform": lambda sum: sum + 13.06182,
        },
    },
    "SexEstimation": {
        "year": 2021,
        "species": "Human",
        "tissue": "Blood",
        "source": "https://bmcgenomics.biomedcentral.com/articles/10.1186/s12864-021-07675-2",
        "output": "Sex",
        "model": {
            "type": "SexEstimationModel",
            "file": "estimateSex.csv",
            "default_imputation": "averaging",
        },
    },
    "LeeRefinedRobust": {
        "year": 2019,
        "species": "Human",
        "tissue": "Placenta",
        "source": "https://www.aging-us.com/article/102049/text",
        "output": "Gestational Age",
        "model": {
            "type": "LinearMethylationModel",
            "file": "LeeRefinedRobust.csv",
            "transform": lambda sum: sum + 30.74966,
        },
    },
    "LeeRobust": {
        "year": 2019,
        "species": "Human",
        "tissue": "Placenta",
        "source": "https://www.aging-us.com/article/102049/text",
        "output": "Gestational Age",
        "model": {
            "type": "LinearMethylationModel",
            "file": "LeeRobust.csv",
            "transform": lambda sum: sum + 24.99772,
        },
    },
    "SmokingMcCartney": {
        "year": 2018,
        "species": "Human",
        "tissue": "Blood",
        "source": "https://www.ncbi.nlm.nih.gov/pmc/articles/PMC6158884/",
        "output": "Smoking Status",
        "model": {"type": "LinearMethylationModel", "file": "Smoking.csv"},
    },
    "DownSyndrome": {
        "year": 2021,
        "species": "Human",
        "tissue": "Blood",
        "source": "https://www.nature.com/articles/s41467-021-21064-z",
        "output": "Down Syndrome Prediction",
        "model": {
            "type": "LinearMethylationModel",
            "file": "down_syndrome.csv",
            "default_imputation": "averaging",
        },
    },
    "DeconvoluteBlood450K": {
        "year": 2024,
        "species": "Human",
        "tissue": "Blood",
        "source": "",
        "output": "Cell Proportions",
        "model": {
            "type": "DeconvolutionModel",
            "file": "450K_reinius_12_reference.csv",
            "platform": "450K",
            "default_imputation": "none",
        },
    },
    "DeconvoluteBloodEPIC": {
        "year": 2024,
        "species": "Human",
        "tissue": "Blood",
        "source": "",
        "output": "Cell Proportions",
        "model": {
            "type": "DeconvolutionModel",
            "file": "EPIC_salas_18_reference.csv",
            "platform": "EPIC",
            "default_imputation": "none",
        },
    },
}


class DeconvolutionModel:
    def __init__(self, reference_file, platform_input):

        # sets reference data
        self.reference = pd.read_csv(
            get_data_file(reference_file), index_col=0
        )

        # sets reference platform type (e.g. "450K")
        self.platform = platform_input

    @classmethod
    def from_definition(cls, clock_definition):

        model_def = clock_definition["model"]

        # calls constructor: "__init__()"
        return cls(model_def["file"], model_def["platform"])

    def predict(self, geo_data):

        # This function computes estimate of cell proportions using quadratic programming
        # inputs:
        # meth_vector: ndarray vector of bulk methylation (length must equal number of rows in deconv_reference)
        # deconv_reference: ndarray matrix with rows corresponding to CpG sites and columns to cell types
        # outputs:
        # cell_prop_estimate.value: ndarray vector of estimated cell type proportions
        def solve_qp(meth_vector, deconv_reference):

            # confirm number of rows in reference equals methylation vector length
            assert deconv_reference.shape[0] == len(meth_vector)

            # define cell proportion variable being solved for
            cell_prop_estimate = cp.Variable(deconv_reference.shape[1])

            # define quadratic objective function
            objective = cp.Minimize(
                cp.norm(deconv_reference @ cell_prop_estimate - meth_vector, 2)
            )

            # define constraints: proportion estimate should be within [0, 1] interval, sum to 1
            constraints = [
                0 <= cell_prop_estimate,
                cell_prop_estimate <= 1,
                cp.sum(cell_prop_estimate) == 1,
            ]

            # create optimization problem
            problem = cp.Problem(objective, constraints)

            # solve optimization problem
            problem.solve(solver=cp.ECOS)

            # return result
            return cell_prop_estimate.value

        # create copy of methylation matrix
        meth = geo_data.dnam.copy()

        # check if missing values are present in methylation matrix
        if any(meth.isna().values.flatten()):
            print(
                "WARNING: methylation data contains missing values. Rows with missing values will be removed"
            )

        # omit rows with missing values
        meth.dropna(inplace=True)

        print(
            "Estimating cell proportions using %r platform reference..."
            % self.platform
        )

        # filter methylation array and reference to include only shared CpGs
        intersecting_cpgs = meth.index.intersection(self.reference.index)

        meth_filtered = meth.loc[intersecting_cpgs]

        reference_filtered = self.reference.loc[intersecting_cpgs]

        # confirm same number of rows in reference and methylation samples
        assert meth_filtered.shape[0] == reference_filtered.shape[0]

        # print number of reference cpgs
        print("Number of CpGs in reference:", len(self.reference.index))
        print(
            "Number of CpGs in methylation data intersecting with reference CpGs:",
            len(intersecting_cpgs),
        )

        # save reference cell types
        cell_types = reference_filtered.columns

        # save methylation sample names
        sample_names = meth_filtered.columns

        # convert methylation samples and reference to ndarray
        meth_filtered_ndarray = meth_filtered.to_numpy()
        reference_filtered_ndarray = reference_filtered.to_numpy()

        print("Running deconvolution...")

        # deconvolve each sample (each column of meth)
        cell_prop = np.apply_along_axis(
            solve_qp,
            axis=0,
            arr=meth_filtered_ndarray,
            deconv_reference=reference_filtered_ndarray,
        )

        # convert cell proportion ndarray to dataframe
        cell_prop_df = pd.DataFrame(cell_prop, columns=sample_names)
        cell_prop_df.index = cell_types

        print("Deconvolution complete!")

        return cell_prop_df

    # returns required methylation sites
    def methylation_sites(self):
        return list(self.reference.index)


class LinearMethylationModel:
    def __init__(
        self, coeffecient_file, transform, preprocess=None, **metadata
    ) -> None:
        self.transform = transform
        self.coefficients = pd.read_csv(
            get_data_file(coeffecient_file), index_col=0
        )
        self.preprocess = preprocess
        self.metadata = metadata

    @classmethod
    def from_definition(cls, clock_definition):
        def no_transform(_):
            return _

        model_def = clock_definition["model"]
        return cls(
            model_def["file"],
            model_def.get("transform", no_transform),
            model_def.get("preprocess", no_transform),
            **{k: v for k, v in clock_definition.items() if k != "model"},
        )

    def predict(self, geo_data):
        dnam_data = self.preprocess(geo_data.dnam)

        # Join the coefficients and dnam_data on the index
        methylation_df = self.coefficients.join(dnam_data, how="inner")

        # Vectorized multiplication: multiply CoefficientTraining with all columns of dnam_data
        result = (
            methylation_df.iloc[:, 1:]
            .multiply(methylation_df["CoefficientTraining"], axis=0)
            .sum(axis=0)
        )

        # Return as a DataFrame
        return result.apply(self.transform).to_frame(name="Predicted")

    def methylation_sites(self):
        return list(self.coefficients.index)


class GrimageModel:
    def __init__(self, coefficient_file, **metadata):
        self.coefficients = pd.read_csv(
            get_data_file(coefficient_file), index_col=0
        )
        self.metadata = metadata

    @classmethod
    def from_definition(cls, clock_definition):
        model_def = clock_definition["model"]
        return cls(
            model_def["file"],
            **{k: v for k, v in clock_definition.items() if k != "model"},
        )

    def predict(self, geo_data):
        if "sex" not in geo_data.metadata or "age" not in geo_data.metadata:
            raise ValueError("Metadata must contain 'sex' and 'age' columns")

        df = geo_data.dnam

        # Transposing metadata so that its structure aligns with dnam (columns as samples)
        transposed_metadata = geo_data.metadata.transpose()

        # Add metadata rows to dnam DataFrame
        df.loc["Age"] = transposed_metadata.loc["age"]
        df.loc["Female"] = transposed_metadata.loc["sex"].apply(
            lambda x: 1 if x == 1 else 0
        )
        df.loc["Intercept"] = 1

        grouped = self.coefficients.groupby("Y.pred")
        all_data = pd.DataFrame()

        for name, group in grouped:
            if name == "COX":
                cox_coefficients = group.set_index("var")["beta"]
            elif name == "transform":
                transform = group.set_index("var")["beta"]
                m_age = transform["m_age"]
                sd_age = transform["sd_age"]
                m_cox = transform["m_cox"]
                sd_cox = transform["sd_cox"]
            else:
                sub_clock_result = self.calculate_sub_clock(df, group)
                all_data[name] = sub_clock_result

        all_data["Age"] = geo_data.metadata["age"]
        all_data["Female"] = geo_data.metadata["sex"].apply(
            lambda x: 1 if x == 1 else 0
        )

        all_data["COX"] = all_data.mul(cox_coefficients).sum(axis=1)
        age_key = "DNAmGrimAge"
        accel_key = "AgeAccelGrim"
        # Calculate DNAmGrimAge
        Y = (all_data["COX"] - m_cox) / sd_cox
        all_data[age_key] = (Y * sd_age) + m_age

        # Calculate AgeAccelGrim
        lm = LinearRegression().fit(
            all_data[["Age"]].values, all_data[age_key].values
        )
        predictions = lm.predict(all_data[["Age"]].values)
        all_data[accel_key] = all_data[age_key] - predictions

        # Drop COX column after computations
        all_data.drop("COX", axis=1, inplace=True)

        return all_data

    def calculate_sub_clock(self, df, coefficients):
        # Filter coefficients for only those present in df
        relevant_coefficients = coefficients[
            coefficients["var"].isin(df.index)
        ]

        # Create a Series from the relevant coefficients, indexed by 'var'
        coefficients_series = relevant_coefficients.set_index("var")["beta"]

        # Align coefficients with df's rows and multiply, then sum across CpG sites for each sample
        result = (
            df.loc[coefficients_series.index]
            .multiply(coefficients_series, axis=0)
            .sum()
        )

        return result

    def rename_columns(self, data, old_names, new_names):
        for old_name, new_name in zip(old_names, new_names):
            data.rename(columns={old_name: new_name}, inplace=True)

    def methylation_sites(self):
        filtered_df = self.coefficients[
            ~self.coefficients.index.isin(["COX", "transform"])
        ]
        unique_vars = set(filtered_df["var"]) - {"Intercept", "Age", "Female"}
        return list(unique_vars)


class SexEstimationModel:
    def __init__(self, coeffecient_file, **metadata):
        self.coefficients = pd.read_csv(
            get_data_file(coeffecient_file), index_col=0, low_memory=False
        )
        self.metadata = metadata

    @classmethod
    def from_definition(cls, clock_definition):
        # Implementation for creating an instance from a definition
        # Adjust this as needed for your specific definition format
        return cls(
            clock_definition["model"]["file"],
            **{k: v for k, v in clock_definition.items() if k != "model"},
        )

    def predict(self, geo_data):
        dnam_data = geo_data.dnam
        # Find the common probes between the reference and the input data
        common_probes = dnam_data.index.intersection(self.coefficients.index)
        reference = self.coefficients.loc[common_probes]
        dnam_data = dnam_data.loc[common_probes]

        # Identify autosomes and calculate mean and standard deviation
        autosomes = reference.loc[~reference["CHR"].isin(["X", "Y"])].index
        d_mean = dnam_data.loc[autosomes].mean(axis=0, skipna=True)
        d_std = dnam_data.loc[autosomes].std(axis=0, skipna=True)

        # Normalize the data using Z-score normalization
        z_data = (
            dnam_data.subtract(d_mean, axis=1).div(d_std, axis=1).fillna(0)
        )

        # Perform the sex prediction for chromosomes X and Y separately
        pred_xy = {}
        for chr in ["X", "Y"]:
            chr_ref = reference.loc[reference["pca"] == chr]
            pred = (z_data.loc[chr_ref.index].T - chr_ref["mean"].values).dot(
                chr_ref["coeff"].values
            )
            pred_xy[chr] = pred

        # Create a DataFrame from the prediction results
        pred_df = pd.DataFrame(pred_xy)

        # Map the prediction results to sex categories
        pred_df["predicted_sex"] = "Female"
        pred_df.loc[
            (pred_df["X"] < 0) & (pred_df["Y"] > 0), "predicted_sex"
        ] = "Male"
        pred_df.loc[
            (pred_df["X"] > 0) & (pred_df["Y"] > 0), "predicted_sex"
        ] = "47,XXY"
        pred_df.loc[
            (pred_df["X"] < 0) & (pred_df["Y"] < 0), "predicted_sex"
        ] = "45,XO"

        return pred_df

    def methylation_sites(self):
        return list(self.coefficients.index)


class ImputationDecorator:
    def __init__(self, clock, imputation_method):
        self.clock = clock
        self.imputation_method = imputation_method

    def predict(self, geo_data):
        # Impute missing values before prediction
        needed_cpgs = self.clock.methylation_sites()
        dnam_data_imputed = self.imputation_method(geo_data.dnam, needed_cpgs)

        return self.clock.predict(
            GeoData(geo_data.metadata, dnam_data_imputed)
        )

    # Forwarding other methods and attributes to the clock
    def __getattr__(self, name):
        return getattr(self.clock, name)


def single_sample_clock(clock_function, data):
    return clock_function(data).iloc[0, 0]



"""
Down Syndrome Epigenetic Plotting
=============================================================

This example loads a DNA Methylation data from GEO with down syndrome metadata and shows how the sameples can be distiguished with epigenetic data
"""

#############################################################################
# First load up some methylation data from GEO using the data library
# ---------------------------------------------------------------------------
from biolearn.data_library import DataLibrary
test_data = DataLibrary().get("GSE52588").load()

######################################################################################
# Now run the down syndrome model on the data to get a score
# ------------------------------------------------------------------------------------

from biolearn.model_gallery import ModelGallery
model = ModelGallery().get("DownSyndrome")
results = model.predict(test_data)


##########################################################################################################
# Finally generate a boxplot to show the predictive power
# --------------------------------------------------------------------------------------------------------
import matplotlib.pyplot as plt
import seaborn as sns

# Set index for merging
results.set_index(results.index.astype(str), inplace=True)
test_data.metadata.set_index(test_data.metadata.index.astype(str), inplace=True)

# Merge data
merged_data = results.join(test_data.metadata)

# Plot settings
category_order = ['healthy', 'Down syndrome']
palette = {'healthy': 'green', 'Down syndrome': 'red'}
title = 'Epigenetic Score'

# Create and configure the plot
plt.figure(figsize=(10, 6))
ax = sns.boxplot(x='disease_state', y='Predicted', data=merged_data, width=0.3, order=category_order)
sns.stripplot(x='disease_state', y='Predicted', data=merged_data, jitter=True, palette=palette, order=category_order, hue='disease_state', dodge=False, legend=False)
ax.set_title(title)
sns.despine()
plt.show()



"""
"Deconvolution Example" 
=============================================================

This example loads blood DNA Methylation data from GEO, estimates the proportion of cell types, and compares them to known cell type proportions. 
"""

#############################################################################
# Load GEO blood methylation dataset measured on EPIC platform
# ---------------------------------------------------------------------------
from biolearn.data_library import DataLibrary

epic_facs_salas_18 = DataLibrary().get("GSE112618").load()

######################################################################################
# Estimate cell proportions using EPIC deconvolution model
# ------------------------------------------------------------------------------------
from biolearn.model_gallery import ModelGallery
gallery = ModelGallery()

epic_facs_cell_prop = gallery.get("DeconvoluteBloodEPIC").predict(epic_facs_salas_18)

######################################################################################
# Compare to known cell proportions measured with FACS (fluorescence-activated cell sorting)
# ------------------------------------------------------------------------------------
import seaborn as sns
import matplotlib.pyplot as plt

sns.set_style("ticks")
plt.figure(figsize = (5, 5))

sns.scatterplot(x = epic_facs_salas_18.metadata.neutrophils_proportion.astype(float), y = epic_facs_cell_prop.loc["neutrophil"])
sns.scatterplot(x = epic_facs_salas_18.metadata.monocytes_proportion.astype(float), y = epic_facs_cell_prop.loc["monocyte"])
sns.scatterplot(x = epic_facs_salas_18.metadata.nk_proportion.astype(float), y = epic_facs_cell_prop.loc["nk_cell"])
sns.scatterplot(x = epic_facs_salas_18.metadata.bcell_proportion.astype(float), y = epic_facs_cell_prop.loc["b_cell"])
sns.scatterplot(x = epic_facs_salas_18.metadata.cd4t_proportion.astype(float), y = epic_facs_cell_prop.loc["cd4_t_cell"])
sns.scatterplot(x = epic_facs_salas_18.metadata.cd8t_proportion.astype(float), y = epic_facs_cell_prop.loc["cd8_t_cell"])

plt.legend(["Neutrophils", "Monocytes", "NK Cells", "B Cells", "CD4 T Cells", "CD8 T Cells"])
plt.xlabel("Known cell proportion")
plt.ylabel("Predicted cell proportion")
plt.title("Deconvolution Validation - FACS")


import pytest
from biolearn.util import get_data_file
from biolearn.model import LinearMethylationModel
from biolearn.model_gallery import ModelGallery


def no_transform(_):
    return _


def test_methylation_model_can_report_sites():
    coefficient_file = get_data_file("Horvath1.csv")
    test_model = LinearMethylationModel(coefficient_file, no_transform)

    some_expected_cpgs = ["cg00945507", "cg04528819", "cg14727952"]
    expected_cpg_count = 353

    result_sites = test_model.methylation_sites()
    assert expected_cpg_count == len(
        result_sites
    ), f"Expected count {expected_cpg_count} but got {len(result_sites)}"
    assert all(
        cpg in result_sites for cpg in some_expected_cpgs
    ), f"Some expected CpGs not found in result"


# Sample Model Data for testing
sample_data = {
    "Horvathv1": {
        "year": 2013,
        "species": "Human",
        "tissue": "Multi-tissue",
        "source": "https://genomebiology.biomedcentral.com/articles/10.1186/gb-2013-14-10-r115",
        "model": {
            "type": "LinearMethylationModel",
            "file": "Horvath1.csv",
            "transform": lambda sum: sum + 0.696,
        },
    },
    "SampleModel2": {
        "year": 2020,
        "species": "Human",
        "tissue": "Brain",
        "source": "https://samplelink.com",
        "model": {
            "type": "LinearMethylationModel",
            "file": "Horvath2.csv",
            "transform": lambda sum: sum + 1.0,
        },
    },
}


def test_init_with_bad_data():
    # This example assumes a bad data structure, adjust accordingly
    bad_data = {"ModelName": "SomeStringInsteadOfDict"}
    with pytest.raises(
        ValueError, match="Expected dictionary for model definition, got"
    ):
        ModelGallery(bad_data)


def test_get_model_by_name():
    gallery = ModelGallery(sample_data)
    model = gallery.get("Horvathv1")
    assert model.metadata["year"] == 2013


def test_get_nonexistent_model_by_name():
    gallery = ModelGallery(sample_data)
    with pytest.raises(KeyError, match="Model not found:"):
        gallery.get("NonExistentModel")


def test_search_with_no_parameters():
    gallery = ModelGallery(sample_data)
    results = gallery.search()
    assert len(results) == len(sample_data)
    assert "Horvathv1" in results
    assert "SampleModel2" in results


def test_search_by_parameters():
    gallery = ModelGallery(sample_data)
    results = gallery.search(species="Human", tissue="Brain")
    assert len(results) == 1
    assert "SampleModel2" in results



"""Functions related to the documentation.

docdict contains the standard documentation entries
used across biolearn.

Entries are listed in alphabetical order.

source: Eric Larson and MNE-python team.
https://github.com/mne-tools/mne-python/blob/main/mne/utils/docs.py
"""

import sys

##############################################################################

# Standard documentation entries
#
# Entries are listed in alphabetical order.
#
docdict = dict()

# annotate
docdict['annotate'] = """
annotate : :obj:`bool`, optional
    If ``annotate`` is ``True``, positions and left/right annotation
    are added to the plot. Default=True."""

# avg_method
docdict['avg_method'] = """
avg_method : {'mean', 'median', 'min', 'max', custom function}, optional
    How to average vertex values to derive the face value:

        - ``mean``: results in smooth boundaries
        - ``median``: results in sharp boundaries
        - ``min`` or ``max``: for sparse matrices
        - ``custom function``: You can also pass a custom function
          which will be executed though :func:`numpy.apply_along_axis`.
          Here is an example of a custom function:

            .. code-block:: python

                def custom_function(vertices):
                    return vertices[0] * vertices[1] * vertices[2]

"""

# ax
docdict['ax'] = """
ax : :class:`~matplotlib.axes.Axes`
    The matplotlib axes in which the plots will be drawn."""

# axes
docdict['axes'] = """
axes : :class:`matplotlib.axes.Axes`, or 4 tuple\
of :obj:`float`: (xmin, ymin, width, height), optional
    The axes, or the coordinates, in matplotlib figure
    space, of the axes used to display the plot.
    If ``None``, the complete figure is used."""

# bg_img
docdict['bg_img'] = """
bg_img : Niimg-like object, optional
    See :ref:`extracting_data`.
    The background image to plot on top of.
"""

# bg_on_data
docdict['bg_on_data'] = r"""
bg_on_data : :obj:`bool`, optional
    If ``True`` and a ``bg_map`` is specified,
    the ``surf_data`` data is multiplied by the background
    image, so that e.g. sulcal depth is jointly visible with ``surf_data``.
    Otherwise, the background image will only be visible where there
    is no surface data (either because ``surf_data`` contains ``nan``\s
    or because is was thresholded).

        .. note::
            This non-uniformly changes the surf_data values according
            to e.g the sulcal depth.

"""

# bg_vmax
docdict['bg_vmax'] = """
bg_vmin : :obj:`float`, optional
    vmax for ``bg_img``."""

# bg_vmin
docdict['bg_vmin'] = """
bg_vmin : :obj:`float`, optional
    vmin for ``bg_img``."""

# black_bg
docdict['black_bg'] = """
black_bg : :obj:`bool`, or 'auto', optional
    If ``True``, the background of the image is set to be black.
    If you wish to save figures with a black background, you
    will need to pass facecolor='k', edgecolor='k'
    to :func:`matplotlib.pyplot.savefig`."""

# border_size
docdict['border_size'] = """
border_size : :obj:`int`, optional
    The size, in :term:`voxel` of the border used on the side of
    the image to determine the value of the background."""

# cbar_tick_format
docdict['cbar_tick_format'] = """
cbar_tick_format : :obj:`str`, optional
    Controls how to format the tick labels of the colorbar.
    Ex: use "%%.2g" to display using scientific notation."""

# classifier_options
svc = "Linear support vector classifier"
logistic = "Logistic regression"
rc = "Ridge classifier"
dc = "Dummy classifier with stratified strategy"

docdict['classifier_options'] = f"""

        - `svc`: :class:`{svc} <sklearn.svm.SVC>` with L2 penalty.
            .. code-block:: python

                svc = LinearSVC(penalty='l2',
                                max_iter=1e4)

        - `svc_l2`: :class:`{svc} <sklearn.svm.SVC>` with L2 penalty.
            .. note::
                Same as option `svc`.

        - `svc_l1`: :class:`{svc} <sklearn.svm.SVC>` with L1 penalty.
            .. code-block:: python

                svc_l1 = LinearSVC(penalty='l1',
                                   dual=False,
                                   max_iter=1e4)

        - `logistic`: \
            :class:`{logistic} <sklearn.linear_model.LogisticRegression>` \
            with L2 penalty.
            .. code-block:: python

                logistic = LogisticRegression(penalty='l2',
                                              solver='liblinear')

        - `logistic_l1`: \
            :class:`{logistic} <sklearn.linear_model.LogisticRegression>` \
            with L1 penalty.
            .. code-block:: python

                logistic_l1 = LogisticRegression(penalty='l1',
                                                 solver='liblinear')

        - `logistic_l2`: \
            :class:`{logistic} <sklearn.linear_model.LogisticRegression>` \
            with L2 penalty
            .. note::
                Same as option `logistic`.

        - `ridge_classifier`: \
            :class:`{rc} <sklearn.linear_model.RidgeClassifierCV>`.
            .. code-block:: python

                ridge_classifier = RidgeClassifierCV()

        - `dummy_classifier`: :class:`{dc} <sklearn.dummy.DummyClassifier>`.
            .. code-block:: python

                dummy = DummyClassifier(strategy='stratified',
                                        random_state=0)

"""

# cmap
docdict['cmap'] = """
cmap : :class:`matplotlib.colors.Colormap`, or :obj:`str`, optional
    The colormap to use. Either a string which is a name of
    a matplotlib colormap, or a matplotlib colormap object."""

# colorbar
docdict['colorbar'] = """
colorbar : :obj:`bool`, optional
    If ``True``, display a colorbar on the right of the plots."""

# connected
docdict['connected'] = """
connected : :obj:`bool`, optional
    If connected is True, only the largest connect component is kept."""


# confounds
docdict['confounds'] = """
confounds : CSV file or array-like, optional
    This parameter is passed to :func:`biolearn.signal.clean`.
    Please see the related documentation for details.
    shape: list of (number of scans, number of confounds)
"""

# cut_coords
docdict['cut_coords'] = """
cut_coords : None, a :obj:`tuple` of :obj:`float`, or :obj:`int`, optional
    The MNI coordinates of the point where the cut is performed.

        - If ``display_mode`` is 'ortho' or 'tiled', this should
          be a 3-tuple: ``(x, y, z)``
        - For ``display_mode == 'x'``, 'y', or 'z', then these are
          the coordinates of each cut in the corresponding direction.
        - If ``None`` is given, the cuts are calculated automatically.
        - If ``display_mode`` is 'mosaic', and the number of cuts is the same
          for all directions, ``cut_coords`` can be specified as an integer.
          It can also be a length 3 tuple specifying the number of cuts for
          every direction if these are different.

        .. note::

            If ``display_mode`` is 'x', 'y' or 'z', ``cut_coords`` can be
            an integer, in which case it specifies the number of
            cuts to perform.

"""

# darkness
docdict['darkness'] = """
darkness : :obj:`float` between 0 and 1, optional
    Specifying the darkness of the background image:

        - '1' indicates that the original values of the background are used
        - '.5' indicates that the background values are reduced by half
          before being applied.

"""

# data_dir
docdict['data_dir'] = """
data_dir : :obj:`pathlib.Path` or :obj:`str`, optional
    Path where data should be downloaded. By default,
    files are downloaded in home directory."""

# detrend
docdict['detrend'] = """
detrend : :obj:`bool`, optional
    Whether to detrend signals or not."""

# dimming factor
docdict['dim'] = """
dim : :obj:`float`, or 'auto', optional
    Dimming factor applied to background image. By default, automatic
    heuristics are applied based upon the background image intensity.
    Accepted float values, where a typical span is between -2 and 2
    (-2 = increase contrast; 2 = decrease contrast), but larger values
    can be used for a more pronounced effect. 0 means no dimming."""

# display_mode
docdict['display_mode'] = """
display_mode : {'ortho', 'tiled', 'mosaic','x',\
'y', 'z', 'yx', 'xz', 'yz'}, optional
    Choose the direction of the cuts:

        - 'x': sagittal
        - 'y': coronal
        - 'z': axial
        - 'ortho': three cuts are performed in orthogonal
          directions
        - 'tiled': three cuts are performed and arranged
          in a 2x2 grid
        - 'mosaic': three cuts are performed along
          multiple rows and columns

    Default='ortho'."""

# draw_cross
docdict['draw_cross'] = """
draw_cross : :obj:`bool`, optional
    If ``draw_cross`` is ``True``, a cross is drawn on the plot to indicate
    the cut position. Default=True."""

# ensure_finite
docdict['ensure_finite'] = """
ensure_finite : :obj:`bool`, optional
    If True, the non-finite values (NANs and infs) found in the data
    will be replaced by zeros."""

# extractor / extract_type
docdict['extractor'] = """
extractor : {'local_regions', 'connected_components'}, optional
    This option can take two values:

        - 'connected_components': each component/region in the image is
          extracted automatically by labelling each region based upon the
          presence of unique features in their respective regions.

        - 'local_regions': each component/region is extracted based on
          their maximum peak value to define a seed marker and then using
          random walker segmentation algorithm on these markers for region
          separation.

    Default='local_regions'."""
docdict['extract_type'] = docdict['extractor']

# figure
docdict['figure'] = """
figure : :obj:`int`, or :class:`matplotlib.figure.Figure`, or None,  optional
    Matplotlib figure used or its number. If ``None`` is given, a
    new figure is created."""

# fsaverage options
docdict['fsaverage_options'] = """

        - 'fsaverage3': the low-resolution fsaverage3 mesh (642 nodes)
        - 'fsaverage4': the low-resolution fsaverage4 mesh (2562 nodes)
        - 'fsaverage5': the low-resolution fsaverage5 mesh (10242 nodes)
        - 'fsaverage6': the medium-resolution fsaverage6 mesh (40962 nodes)
        - 'fsaverage7': same as 'fsaverage'
        - 'fsaverage': the high-resolution fsaverage mesh (163842 nodes)

            .. note::
                The high-resolution fsaverage will result in more computation
                time and memory usage

"""

# fwhm
docdict['fwhm'] = """
fwhm : scalar, :class:`numpy.ndarray`, or :obj:`tuple`, or :obj:`list`,\
or 'fast' or None, optional
    Smoothing strength, as a :term:`full-width at half maximum<FWHM>`,
    in millimeters:

        - If a nonzero scalar is given, width is identical in all 3 directions.
        - If a :class:`numpy.ndarray`, :obj:`tuple`, or :obj:`list` is given,
          it must have 3 elements, giving the :term:`FWHM` along each axis.
          If any of the elements is zero or None, smoothing is not performed
          along that axis.
        - If `fwhm='fast'`, a fast smoothing will be performed with a filter
          [0.2, 1, 0.2] in each direction and a normalisation to preserve the
          local average value.
        - If ``fwhm`` is None, no filtering is performed (useful when just
          removal of non-finite values is needed).

    .. note::

        In corner case situations, `fwhm` is simply kept to None when `fwhm`
        is specified as `fwhm=0`.

"""

# hemi
docdict['hemi'] = """
hemi : {'left', 'right'}, optional
    Hemisphere to display. Default='left'."""

# hemispheres
docdict['hemispheres'] = """
hemispheres : :obj:`list` of :obj:`str`, optional
    Hemispheres to display. Default=['left', 'right']."""


# high_pass
docdict['high_pass'] = """
high_pass : :obj:`float`, optional
    High cutoff frequency in Hertz. If specified, signals below this
    frequency will be filtered out. Default=None."""

# hrf_model
docdict['hrf_model'] = """
hrf_model : :obj:`str`, function, list of functions, or None
    This parameter defines the :term:`HRF` model to be used.
    It can be a string if you are passing the name of a model
    implemented in biolearn.
    Valid names are:

        - 'spm': This is the :term:`HRF` model used in :term:`SPM`.
          See :func:`biolearn.glm.first_level.spm_hrf`.
        - 'spm + derivative': SPM model plus its time derivative.
          This gives 2 regressors.
          See :func:`biolearn.glm.first_level.spm_hrf`, and
          :func:`biolearn.glm.first_level.spm_time_derivative`.
        - 'spm + derivative + dispersion': Idem, plus dispersion
          derivative. This gives 3 regressors.
          See :func:`biolearn.glm.first_level.spm_hrf`,
          :func:`biolearn.glm.first_level.spm_time_derivative`,
          and :func:`biolearn.glm.first_level.spm_dispersion_derivative`.
        - 'glover': This corresponds to the Glover :term:`HRF`.
          See :func:`biolearn.glm.first_level.glover_hrf`.
        - 'glover + derivative': The Glover :term:`HRF` + time
          derivative. This gives 2 regressors.
          See :func:`biolearn.glm.first_level.glover_hrf`, and
          :func:`biolearn.glm.first_level.glover_time_derivative`.
        - 'glover + derivative + dispersion': Idem, plus dispersion
          derivative. This gives 3 regressors.
          See :func:`biolearn.glm.first_level.glover_hrf`,
          :func:`biolearn.glm.first_level.glover_time_derivative`, and
          :func:`biolearn.glm.first_level.glover_dispersion_derivative`.
        - 'fir': Finite impulse response basis. This is a set of
          delayed dirac models.

    It can also be a custom model. In this case, a function should
    be provided for each regressor. Each function should behave as the
    other models implemented within biolearn. That is, it should take
    both `t_r` and `oversampling` as inputs and return a sample numpy
    array of appropriate shape.

    .. note::
        It is expected that `spm` standard and `glover` models would
        not yield large differences in most cases.

    .. note::
        In case of `glover` and `spm` models, the derived regressors
        are orthogonalized wrt the main one.

"""

# img
docdict['img'] = """
img : Niimg-like object
    See :ref:`extracting_data`.
"""

# imgs
docdict['imgs'] = """
imgs : :obj:`list` of Niimg-like objects
    See :ref:`extracting_data`.
"""

# legacy_format
docdict['legacy_format'] = """
legacy_format : :obj:`bool`, optional
    If set to ``True``, the fetcher will return recarrays. Otherwise,
    it will return pandas dataframes.
    Default=True."""

# linewidth
docdict['linewidths'] = """
linewidths : :obj:`float`, optional
    Set the boundary thickness of the contours.
    Only reflects when ``view_type=contours``."""

# low_pass
docdict['low_pass'] = """
low_pass : :obj:`float` or None, optional
    Low cutoff frequency in Hertz. If specified, signals above this
    frequency will be filtered out. If None, no low-pass filtering
    will be performed. Default=None."""

# lower_cutoff
docdict['lower_cutoff'] = """
lower_cutoff : :obj:`float`, optional
    Lower fraction of the histogram to be discarded."""

# mask_img
docdict['mask_img'] = """
mask_img : Niimg-like object
    Object used for masking the data."""

# mask_strategy
docdict["mask_strategy"] = """
mask_strategy : {'background', 'epi', 'whole-brain-template',\
'gm-template', 'wm-template'}, optional
    The strategy used to compute the mask:

        - 'background': Use this option if your images present
          a clear homogeneous background.
        - 'epi': Use this option if your images are raw EPI images
        - 'whole-brain-template': This will extract the whole-brain
          part of your data by resampling the MNI152 brain mask for
          your data's field of view.

            .. note::
                This option is equivalent to the previous 'template' option
                which is now deprecated.

        - 'gm-template': This will extract the gray matter part of your
          data by resampling the corresponding MNI152 template for your
          data's field of view.

            .. versionadded:: 0.8.1

        - 'wm-template': This will extract the white matter part of your
          data by resampling the corresponding MNI152 template for your
          data's field of view.

            .. versionadded:: 0.8.1

"""

# mask_type
docdict['mask_type'] = """
mask_type : {'whole-brain', 'gm', 'wm'}, optional
    Type of mask to be computed:

        - 'whole-brain': Computes the whole-brain mask.
        - 'gm': Computes the grey-matter mask.
        - 'wm': Computes the white-matter mask.

    Default = 'whole-brain'.

"""

# kwargs for Maskers
docdict['masker_kwargs'] = """
kwargs : dict
    Keyword arguments to be passed to functions called within the masker.
    Kwargs prefixed with ``'clean__'`` will be passed to
    :func:`~biolearn.signal.clean`.
    Within :func:`~biolearn.signal.clean`, kwargs prefixed with
    ``'butterworth__'`` will be passed to the Butterworth filter
    (i.e., ``clean__butterworth__``).
"""

# memory
docdict['memory'] = """
memory : instance of :class:`joblib.Memory`, :obj:`str`, or \
:class:`pathlib.Path`
    Used to cache the masking process.
    By default, no caching is done. If a :obj:`str` is given, it is the
    path to the caching directory."""

# memory_level
memory_level = """
memory_level : :obj:`int`, optional.
    Rough estimator of the amount of memory used by caching. Higher value
    means more memory for caching. Zero means no caching.
    Default={}."""
docdict['memory_level'] = memory_level.format(0)
docdict['memory_level1'] = memory_level.format(1)

# n_jobs
n_jobs = """
n_jobs : :obj:`int`, optional.
    The number of CPUs to use to do the computation. -1 means 'all CPUs'.
    Default={}."""
docdict['n_jobs'] = n_jobs.format("1")
docdict['n_jobs_all'] = n_jobs.format("-1")

# n_parcels
docdict['n_parcels'] = """
n_parcels : :obj:`int`, optional
    Number of parcels to divide the data into.
    Default=50."""

# opening
docdict['opening'] = """
opening : :obj:`bool` or :obj:`int`, optional
    This parameter determines whether a morphological
    :term:`opening<Opening>` is performed, to keep only large structures.
    This step is useful to remove parts of the skull that might have been
    included. ``opening`` can be:

        - A boolean : If False, no :term:`opening<Opening>` is performed.
          If True, it is equivalent to ``opening=1``.
        - An integer `n`: The :term:`opening<Opening>` is performed via `n`
          :term:`erosions<Erosion>` (see :func:`scipy.ndimage.binary_erosion`).
          The largest connected component is then estimated if ``connected`` is
          set to True, and 2`n` :term:`dilation<Dilation>` operations are
          performed (see :func:`scipy.ndimage.binary_dilation`) followed by
          `n` :term:`erosions<Erosion>`. This corresponds to 1
          :term:`opening<Opening>` operation of order `n` followed by a
          :term:`closing<Closing>` operator of order `n`.

    .. note::

        Turning off :term:`opening<Opening>` (``opening=False``) will also
        prevent any smoothing applied to the image during the mask computation.

"""

# output_file
docdict['output_file'] = """
output_file : :obj:`str`, or None, optional
    The name of an image file to export the plot to. Valid extensions
    are .png, .pdf, .svg. If ``output_file`` is not None, the plot
    is saved to a file, and the display is closed."""

# random_state
docdict['random_state'] = """
random_state : :obj:`int` or RandomState, optional
    Pseudo-random number generator state used for random sampling."""

# regressor_options
docdict['regressor_options'] = """

        - `ridge`: \
            :class:`{Ridge regression} <sklearn.linear_model.RidgeCV>`.
            .. code-block:: python

                ridge = RidgeCV()

        - `ridge_regressor`: \
            :class:`{Ridge regression} <sklearn.linear_model.RidgeCV>`.
            .. note::
                Same option as `ridge`.

        - `svr`: :class:`{Support vector regression} <sklearn.svm.SVR>`.
            .. code-block:: python

                svr = SVR(kernel='linear',
                          max_iter=1e4)

        - `dummy_regressor`: \
            :class:`{Dummy regressor} <sklearn.dummy.DummyRegressor>`.
            .. code-block:: python

                dummy = DummyRegressor(strategy='mean')

"""

# resampling_interpolation
docdict['resampling_interpolation'] = """
resampling_interpolation : :obj:`str`, optional
    Interpolation to use when resampling the image to
    the destination space. Can be:

        - "continuous": use 3rd-order spline interpolation
        - "nearest": use nearest-neighbor mapping.

            .. note::

                "nearest" is faster but can be noisier in some cases.

"""

# resume
docdict['resume'] = """
resume : :obj:`bool`, optional
    Whether to resume download of a partly-downloaded file.
    Default=True."""

# sample_mask
docdict['sample_mask'] = """
sample_mask : Any type compatible with numpy-array indexing, optional
    shape: (number of scans - number of volumes removed, )
    Masks the niimgs along time/fourth dimension to perform scrubbing
    (remove volumes with high motion) and/or non-steady-state volumes.
    This parameter is passed to :func:`biolearn.signal.clean`.
"""

# second_level_contrast
docdict['second_level_contrast'] = """
second_level_contrast : :obj:`str` or :class:`numpy.ndarray` of shape\
(n_col), optional
    Where ``n_col`` is the number of columns of the design matrix.
    The string can be a formula compatible with :meth:`pandas.DataFrame.eval`.
    Basically one can use the name of the conditions as they appear in the
    design matrix of the fitted model combined with operators +- and combined
    with numbers with operators +-`*`/. The default (None) is accepted if the
    design matrix has a single column, in which case the only possible
    contrast array((1)) is applied; when the design matrix has multiple
    columns, an error is raised."""

# second_level_input
docdict['second_level_input'] = """
second_level_input : :obj:`list` of \
:class:`~biolearn.glm.first_level.FirstLevelModel` objects \
or :class:`pandas.DataFrame` or :obj:`list` of Niimg-like objects.

    - Giving :class:`~biolearn.glm.first_level.FirstLevelModel` objects
      will allow to easily compute the second level contrast of arbitrary first
      level contrasts thanks to the ``first_level_contrast`` argument of
      :meth:`~biolearn.glm.first_level.FirstLevelModel.compute_contrast`.
      Effect size images will be computed for each model to contrast at
      the second level.
    - If a :class:`~pandas.DataFrame`, then it has to contain
      ``subject_label``, ``map_name`` and ``effects_map_path``. It can
      contain multiple maps that would be selected during contrast estimation
      with the argument ``first_level_contrast`` of
      :meth:`~biolearn.glm.first_level.FirstLevelModel.compute_contrast`.
      The :class:`~pandas.DataFrame` will be sorted based on
      the ``subject_label`` column to avoid order inconsistencies when
      extracting the maps. So the rows of the automatically computed
      design matrix, if not provided, will correspond to the sorted
      ``subject_label`` column.
    - If a :obj:`list` of Niimg-like objects then this is taken
      literally as Y for the model fit and ``design_matrix`` must be
      provided.

"""

# smoothing_fwhm
docdict['smoothing_fwhm'] = """
smoothing_fwhm : :obj:`float`, optional.
    If ``smoothing_fwhm`` is not ``None``, it gives
    the :term:`full-width at half maximum<FWHM>` in millimeters
    of the spatial smoothing to apply to the signal."""

# standardize
standardize = """
standardize : :obj:`bool`, optional.
    If ``standardize`` is True, the data are centered and normed:
    their mean is put to 0 and their variance is put to 1 in the
    time dimension.
    Default={}."""
docdict['standardize'] = standardize.format('True')
docdict['standardize_false'] = standardize.format('False')

# standardize as used within maskers module
docdict["standardize_maskers"] = """
standardize : {'zscore_sample', 'zscore', 'psc', True, False}, default=False
    Strategy to standardize the signal:

        - 'zscore_sample': The signal is z-scored. Timeseries are shifted
          to zero mean and scaled to unit variance. Uses sample std.
        - 'zscore': The signal is z-scored. Timeseries are shifted
          to zero mean and scaled to unit variance. Uses population std
          by calling default :obj:`numpy.std` with N - ``ddof=0``.
        - 'psc':  Timeseries are shifted to zero mean value and scaled
          to percent signal change (as compared to original mean signal).
        - True: The signal is z-scored (same as option `zscore`).
          Timeseries are shifted to zero mean and scaled to unit variance.
        - False: Do not standardize the data.
"""

# standardize_confounds
docdict['standardize_confounds'] = """
standardize_confounds : :obj:`bool`, optional
    If set to True, the confounds are z-scored: their mean is put
    to 0 and their variance to 1 in the time dimension.
    Default=True."""

# symmetric_cbar
docdict['symmetric_cbar'] = """
symmetric_cbar : :obj:`bool`, or 'auto', optional
    Specifies whether the colorbar should range from ``-vmax`` to ``vmax``
    or from ``vmin`` to ``vmax``. Setting to 'auto' will select the latter
    if the range of the whole image is either positive or negative.

    .. note::

        The colormap will always range from ``-vmax`` to ``vmax``.

"""

# t_r
docdict['t_r'] = """
t_r : :obj:`float` or None, optional
    Repetition time, in seconds (sampling period).
    Set to ``None`` if not provided.
    Default=None."""

# target_affine
docdict['target_affine'] = """
target_affine : :class:`numpy.ndarray`, optional.
    If specified, the image is resampled corresponding to this new affine.
    ``target_affine`` can be a 3x3 or a 4x4 matrix.
    Default=None."""

# target_shape
docdict['target_shape'] = """
target_shape : :obj:`tuple` or :obj:`list`, optional.
    If specified, the image will be resized to match this new shape.
    ``len(target_shape)`` must be equal to 3.

    .. note::
        If ``target_shape`` is specified, a ``target_affine`` of shape
        ``(4, 4)`` must also be given.

    Default=None."""

# templateflow
docdict["templateflow"] = """
    The default template of :term:`fMRIPrep` is the asymmetrical ICBM152 2009,
    release c (MNI152NLin2009cSAsym). The biolearn template is asymmetrical
    ICBM152 2009, release a. If you wish to use the exact same release as
    :term:`fMRIPrep`, please refer to TemplateFlow
    (https://www.templateflow.org/).
"""

# threshold
docdict['threshold'] = """
threshold : a number, None, or 'auto', optional
    If ``None`` is given, the image is not thresholded.
    If a number is given, it is used to threshold the image:
    values below the threshold (in absolute value) are plotted
    as transparent. If 'auto' is given, the threshold is determined
    magically by analysis of the image.
"""

# title
docdict['title'] = """
title : :obj:`str`, or None, optional
    The title displayed on the figure.
    Default=None."""

# upper_cutoff
docdict['upper_cutoff'] = """
upper_cutoff : :obj:`float`, optional
    Upper fraction of the histogram to be discarded."""

# url
docdict['url'] = """
url : :obj:`str`, optional
    URL of file to download.
    Override download URL. Used for test only (or if you
    setup a mirror of the data).
    Default=None."""

# verbose
verbose = """
verbose : :obj:`int`, optional
    Verbosity level (0 means no message).
    Default={}."""
docdict['verbose'] = verbose.format(1)
docdict['verbose0'] = verbose.format(0)

# view
docdict['view'] = """
view : :obj:`str` or a pair of :obj:`float`, optional.
    If a string, must be in {'lateral', 'medial', 'dorsal', 'ventral',\
    'anterior', 'posterior'}.
    If a sequence, must be a pair (elev, azim) of float
    angles in degrees that will manually set a custom view.
    E.g., view=[270.0, 90.0] or view=(0.0, -180.0).
    View of the surface that is rendered.
    Default='lateral'.

"""

# vmax
docdict['vmax'] = """
vmax : :obj:`float`, optional
    Upper bound of the colormap. If ``None``, the max of the image is used.
    Passed to :func:`matplotlib.pyplot.imshow`.
"""

# vmin
docdict['vmin'] = """
vmin : :obj:`float`, optional
    Lower bound of the colormap. If ``None``, the min of the image is used.
    Passed to :func:`matplotlib.pyplot.imshow`.
"""

##############################################################################

docdict_indented = {}


def _indentcount_lines(lines):
    """Minimum indent for all lines in line list.

    >>> lines = [' one', '  two', '   three']
    >>> _indentcount_lines(lines)
    1
    >>> lines = []
    >>> _indentcount_lines(lines)
    0
    >>> lines = [' one']
    >>> _indentcount_lines(lines)
    1
    >>> _indentcount_lines(['    '])
    0

    """
    indentno = sys.maxsize
    for line in lines:
        stripped = line.lstrip()
        if stripped:
            indentno = min(indentno, len(line) - len(stripped))
    if indentno == sys.maxsize:
        return 0
    return indentno


def fill_doc(f):
    """Fill a docstring with docdict entries.

    Parameters
    ----------
    f : callable
        The function to fill the docstring of. Will be modified in place.

    Returns
    -------
    f : callable
        The function, potentially with an updated ``__doc__``.

    """
    docstring = f.__doc__
    if not docstring:
        return f
    lines = docstring.splitlines()
    # Find the minimum indent of the main docstring, after first line
    if len(lines) < 2:
        icount = 0
    else:
        icount = _indentcount_lines(lines[1:])
    # Insert this indent to dictionary docstrings
    try:
        indented = docdict_indented[icount]
    except KeyError:
        indent = ' ' * icount
        docdict_indented[icount] = indented = {}
        for name, dstr in docdict.items():
            lines = dstr.splitlines()
            try:
                newlines = [lines[0]]
                for line in lines[1:]:
                    newlines.append(indent + line)
                indented[name] = '\n'.join(newlines)
            except IndexError:
                indented[name] = dstr
    try:
        f.__doc__ = docstring % indented
    except (TypeError, ValueError, KeyError) as exp:
        funcname = f.__name__
        funcname = docstring.split('\n')[0] if funcname is None else funcname
        raise RuntimeError('Error documenting %s:\n%s'
                           % (funcname, str(exp)))
    return f



import pandas as pd
import numpy as np


# Biomarker parameters from the paper https://www.ncbi.nlm.nih.gov/pmc/articles/PMC5940111/
def phenotypic_age(df):
    pheno_coefs = {
        "age": 0.0804,
        "albumin": -0.034,
        "creatinine": 0.0095,
        "glucose": 0.1953,
        "c_reactive_protein": 0.0954,
        "lymphocyte_percent": -0.012,
        "mean_cell_volume": 0.0268,
        "red_blood_cell_distribution_width": 0.3356,
        "alkaline_phosphate": 0.00188,
        "white_blood_cell_count": 0.0554,
    }

    constant = -19.9067
    gamma = 0.0077
    cs = [141.50225, -0.00553, 0.090165]
    df["c_reactive_protein"] = np.log(df["c_reactive_protein"])
    df["pheno"] = df.apply(
        lambda x: sum([x[c] * pheno_coefs[c] for c in pheno_coefs.keys()]),
        axis=1,
    )
    df["mortality_score"] = 1 - np.exp(
        (-np.exp(constant + df["pheno"]) * ((np.exp(120 * gamma) - 1) / gamma))
    )
    df["phenotypic_age"] = (
        cs[0] + np.log(cs[1] * np.log(1.00001 - df["mortality_score"])) / cs[2]
    )
    return df["phenotypic_age"]



"""
"Epigenetic Clocks" in GEO Data
=============================================================

This example loads a DNA Methylation data from GEO, calculates multiple epigenetic clocks, and plots them against chronological age. 
"""

#############################################################################
# First load up some methylation data from GEO using the data library
# ---------------------------------------------------------------------------
from biolearn.data_library import DataLibrary
#Load up GSE41169 blood DNAm data
data_source = DataLibrary().get("GSE41169")
data=data_source.load()


######################################################################################
# Now run three different clocks on the dataset to produce epigenetic clock ages
# ------------------------------------------------------------------------------------

from biolearn.model_gallery import ModelGallery
gallery = ModelGallery()
#Note that by default clocks will impute missing data.
#To change this behavior set the imputation= parameter when getting the clock
horvath_results = gallery.get("Horvathv1").predict(data)
hannum_results = gallery.get("Hannum").predict(data)
phenoage_results = gallery.get("PhenoAge").predict(data)



##########################################################################################################
# Finally extract the age data from the metadata from GEO and plot the results using seaborn
# --------------------------------------------------------------------------------------------------------
import seaborn as sn
import pandas as pd

actual_age = data.metadata['age']
plot_data = pd.DataFrame({
    'Horvath': horvath_results['Predicted'],
    'Hannum': hannum_results['Predicted'],
    'PhenoAge': phenoage_results['Predicted'],
    "Age": actual_age
})
plot_data.index=plot_data['Age']

sn.relplot(data=plot_data, kind="scatter");


"""
Building a competition submission using an existing model
=============================================================

This example shows you how to generate a submission for the community warmup using an existing model
"""

#############################################################################
# Loading up the data for the competition
# ---------------------------------------
from biolearn.data_library import GeoData

#Download the data file for the warmup challenge linked here https://www.synapse.org/#!Synapse:syn52966292/wiki/625231
DOWNLOADED_DATA_FILE_PATH="ADD YOUR PATH HERE"
challenge_data = GeoData.from_methylation_matrix(DOWNLOADED_DATA_FILE_PATH)
challenge_data

#############################################################################
# Use the Lin model to predict the age
# ------------------------------------------------------
from biolearn.model_gallery import ModelGallery

#Lin scored the best so lets use that
results = ModelGallery().get("Lin").predict(challenge_data)
results

############################################################################################################################
# Save a csv in the correct format for submission
# --------------------------------------------------------------------------------------------------------------------------
#Now we save this to a csv in the expected format. 
submission = results.rename_axis("sampleId")
submission = submission.rename(columns={"Predicted": "predictedAge"})
submission.to_csv("lin_submission.csv")
#You can submit this file to the community warmup challenge by following the instructions below
submission


import pytest
from math import isclose
import pandas as pd
import numpy as np
from biolearn import model
from biolearn.util import (
    get_test_data_file,
    load_test_data_file,
)
from biolearn.data_library import GeoData
import pickle


sample_inputs = load_test_data_file("external/DNAmTestSet.csv")
sample_metadata = load_test_data_file("external/testset_metadata.csv")


@pytest.mark.parametrize(
    "model_name, model_entry", model.model_definitions.items()
)
def test_models(model_name, model_entry):
    # Skip models not implemented
    if model_entry["model"]["type"] == "NotImplemented":
        pytest.skip(
            f"Model type 'NotImplemented' for {model_name} - skipping test"
        )

    test_data = GeoData(sample_metadata, sample_inputs)

    # Check if the model class exists
    try:
        model_class = getattr(model, model_entry["model"]["type"])
    except AttributeError:
        pytest.fail(
            f"Model class '{model_entry['model']['type']}' not found in biolearn.model"
        )

    # Instantiate the model
    test_model = model_class.from_definition(model_entry)

    actual_results = test_model.predict(test_data.copy()).sort_index()

    # Load the expected results
    expected_results = load_test_data_file(
        f"expected_model_outputs/{model_name}.csv"
    ).sort_index()

    # Assertions and discrepancy checks
    assert len(expected_results) == len(
        actual_results
    ), f"For {model_name}: DataFrames do not have the same length"

    discrepancies = []
    for idx in expected_results.index:
        for col in expected_results.columns:
            expected_val = expected_results.loc[idx, col]
            actual_val = actual_results.loc[idx, col]

            if isinstance(expected_val, float) and not isclose(
                actual_val, expected_val, abs_tol=1e-5
            ):
                discrepancies.append((idx, col, expected_val, actual_val))
            elif (
                not isinstance(expected_val, float)
                and expected_val != actual_val
            ):
                print(type(expected_val))
                discrepancies.append((idx, col, expected_val, actual_val))

    assert not discrepancies, "\n".join(
        [
            f"For {model_name}: Discrepancy at index {idx}, column {col}: expected {expected}, but got {actual}"
            for idx, col, expected, actual in discrepancies
        ]
    )


def test_dunedin_pace_normalization():
    actual = model.dunedin_pace_normalization(sample_inputs)
    data_file_path = get_test_data_file("pace_normalized.pkl")
    with open(data_file_path, "rb") as file:
        expected = pickle.load(file)

    # Finding mismatches based on tolerance
    mask = np.abs(actual - expected) > 0.00000001
    mismatches = actual[mask].stack()

    total_mismatches = mismatches.size
    percentage_mismatched = (total_mismatches / actual.size) * 100

    # Display the mismatches
    for idx, value in enumerate(mismatches.items()):
        if idx == 100:
            break
        print(
            f"Location: {value[0]}, Actual: {actual.at[value[0]]}, Expected: {expected.at[value[0]]}"
        )

    print(
        f"Total mismatches: {total_mismatches} ({percentage_mismatched:.2f}%)"
    )

    # Your actual assertion can be here based on your needs, e.g.,
    assert (
        total_mismatches == 0
    ), "Dataframes are not equal within the given tolerance."


# Run the test
if __name__ == "__main__":
    pytest.main([__file__])



"""
Performing custom imputations
=============================================================

This example demonstrates the ease of custom imputation using the biolearn library.
"""

################################################################################
# Load methylation data from a GEO dataset to use as a reference for imputation
# ------------------------------------------------------------------------------

from biolearn.data_library import DataLibrary, GeoData

# Load the reference dataset and compute averages for imputation
reference_dataset = DataLibrary().get("GSE40279").load()
reference_averages = reference_dataset.dnam.mean(axis=1)

#############################################################################
# Load up a target dataset and run the imputation
# --------------------------------------------------------------------------

# Load a second dataset for imputation
target_dataset = DataLibrary().get("GSE51057").load()

# Perform imputation using the reference averages
from biolearn.imputation import impute_from_standard
imputed_data = impute_from_standard(target_dataset.dnam, reference_averages)
imputed_dataset = GeoData(target_dataset.metadata, imputed_data)

#############################################################################
# Now run clock predictions on the dataset before and after
# --------------------------------------------------------------------------

# Using a model from the gallery, compare the epigenetic age before and after imputation
from biolearn.model_gallery import ModelGallery
clock_model = ModelGallery().get("Horvathv1")

original_age_predictions = clock_model.predict(target_dataset)
imputed_age_predictions = clock_model.predict(imputed_dataset)

#############################################################################
# Visualize the comparison of age predictions
# --------------------------------------------------------------------------

import seaborn as sns
import pandas as pd
import matplotlib.pyplot as plt

# Prepare data for visualization
comparison_data = pd.DataFrame({
    'Original': original_age_predictions['Predicted'],
    'Imputed': imputed_age_predictions['Predicted']
})

# Create a scatter plot to compare the results
plt.figure(figsize=(8, 6))
sns.scatterplot(x='Original', y='Imputed', data=comparison_data)
plt.title("Comparison of Epigenetic Age Predictions: Original vs Imputed")
plt.xlabel("Original Age Predictions")
plt.ylabel("Imputed Age Predictions")
plt.grid(True)
plt.show()


#############################################################################
# You can also build an imputation decorator to bundle with the clock
# --------------------------------------------------------------------------

from biolearn.model import ImputationDecorator

# Define a custom imputation method using reference averages
def custom_impute_method(dnam_data, needed_cpgs):
    return impute_from_standard(dnam_data, reference_averages, needed_cpgs)

# Wrap the clock model with the imputation decorator
decorated_clock = ImputationDecorator(clock_model, custom_impute_method)

#############################################################################
# You get the same results as running the function directly on your dataset
# --------------------------------------------------------------------------

# Predict epigenetic age using the decorated clock (with imputation)
decorated_clock_predictions = decorated_clock.predict(target_dataset)

# Verify that the results are the same as the direct imputation approach
same_results = all(decorated_clock_predictions == imputed_age_predictions)
print(same_results)



from biolearn import load
import pytest


load_columns = ["sex", "age", "glucose", "is_dead", "months_until_death"]


def test_fhs_columns():
    df = load.load_fhs()
    verify_expected_columns(df)


def test_nhanes_columns():
    df = load.load_nhanes(2010)
    verify_expected_columns(df)


def test_can_load_nhanes_2012():
    df = load.load_nhanes(2012)


def test_expected_error_when_loading_unsupported_year_nhanes():
    with pytest.raises(ValueError):
        df = load.load_nhanes(1913)


def verify_expected_columns(df):
    actual_columns = set(df.columns.to_list())
    missing_columns = set(load_columns) - actual_columns
    extra_columns = actual_columns - set(load_columns)
    assert (
        len(missing_columns) == 0
    ), f"Missing expected columns: {missing_columns} \n Found extra columns: {extra_columns}"


# Run the test
if __name__ == "__main__":
    pytest.main([__file__])



import pandas as pd
from biolearn.util import cached_download

MG_PER_DL_TO_MMOL_PER_L = 0.05551


def load_fhs():
    """
    Loads data from the Framingham Heart Study

    Returns
    -------
    df: Pandas.Dataframe
        A pandas dataframe where each row represents an individual and each column represents a measurement about that individual
    """
    public_link = "https://raw.githubusercontent.com/singator/bdah/master/data/frmgham2.csv"
    df = pd.read_csv(
        public_link,
        index_col=0,
        usecols=[
            "RANDID",
            "PERIOD",
            "AGE",
            "SEX",
            "DEATH",
            "TIMEDTH",
            "GLUCOSE",
        ],
    )
    df = df[df["PERIOD"] == 1].drop("PERIOD", axis=1)
    df["TIMEDTH"] = df["TIMEDTH"] / 30.437  # days to months
    df.index.name = "id"

    df = df.rename(
        {
            "AGE": "age",
            "SEX": "sex",
            "GLUCOSE": "glucose",
            "DEATH": "is_dead",
            "TIMEDTH": "months_until_death",
        },
        axis=1,
    )

    # standardize glucose units
    df["glucose"] = df["glucose"].apply(lambda g: g * MG_PER_DL_TO_MMOL_PER_L)

    return df


def load_nhanes(year):
    """Loads data from the National Health and Nutrition Examination Survey

    Parameters
    ----------
    year : number
        A year number for which to load data. Not that NHANES data comes in two year groupings and the year passed in should be the later year.
        Supported inputs are 2010 and 2012

    Returns
    -------
    df: Pandas.Dataframe
        A pandas dataframe where each row represents an individual and each column represents a measurement about that individual
    """
    cbc_sub = [
        "LBXRDW",
        "LBXWBCSI",
        "LBXLYPCT",
        "LBXMCVSI",
        "LBDLYMNO",
        "LBXRBCSI",
        "LBXHGB",
        "LBXPLTSI",
        "LBXMCHSI",
        "LBXBAPCT",
    ]
    known_nhanes_year_suffix = {2010: "F", 2012: "G"}
    if not year in known_nhanes_year_suffix:
        raise ValueError(
            f"Unknown year {year}. Can only load for known available years {known_nhanes_year_suffix.keys}"
        )
    suffix = known_nhanes_year_suffix[year]
    dem_file = cached_download(
        f"https://wwwn.cdc.gov/Nchs/Nhanes/{year-1}-{year}/DEMO_{suffix}.XPT"
    )
    gluc_file = cached_download(
        f"https://wwwn.cdc.gov/Nchs/Nhanes/{year-1}-{year}/GLU_{suffix}.XPT"
    )
    cbc_file = cached_download(
        f"https://wwwn.cdc.gov/Nchs/Nhanes/{year-1}-{year}/CBC_{suffix}.XPT"
    )
    bioc_file = cached_download(
        f"https://wwwn.cdc.gov/Nchs/Nhanes/{year-1}-{year}/BIOPRO_{suffix}.XPT"
    )
    mortality_file = cached_download(
        f"https://ftp.cdc.gov/pub/Health_Statistics/NCHS/datalinkage/linked_mortality/NHANES_{year-1}_{year}_MORT_2019_PUBLIC.dat"
    )
    crp_file = cached_download(
        f"https://wwwn.cdc.gov/Nchs/Nhanes/{year-1}-{year}/CRP_{suffix}.XPT"
    )
    hdl_file = cached_download(
        f"https://wwwn.cdc.gov/Nchs/Nhanes/{year-1}-{year}/HDL_{suffix}.XPT"
    )
    dem = pd.read_sas(dem_file, index="SEQN")[["RIAGENDR", "RIDAGEYR"]]
    dem.index = dem.index.astype(int)
    gluc = pd.read_sas(gluc_file, index="SEQN")["LBDGLUSI"]
    gluc.index = gluc.index.astype(int)
    cbc = pd.read_sas(cbc_file, index="SEQN")[cbc_sub]
    cbc.index = cbc.index.astype(int)
    hdl = pd.read_sas(hdl_file, index="SEQN")["LBDHDDSI"]
    hdl.index = hdl.index.astype(int)
    # clumsy hack since 2012 doesn't have the CRP data. Will remove pending refactor of loading code
    if year == 2010:
        crp = pd.read_sas(crp_file, index="SEQN")["LBXCRP"]
        crp.index = crp.index.astype(int)
    bioc = pd.read_sas(bioc_file, index="SEQN")[
        ["LBDSALSI", "LBDSCRSI", "LBXSAPSI"]
    ]
    bioc.index = bioc.index.astype(int)
    mort = pd.read_fwf(
        mortality_file,
        index_col=0,
        header=None,
        widths=[14, 1, 1, 3, 1, 1, 1, 4, 8, 8, 3, 3],
    )
    mort.index = mort.index.rename("SEQN")
    dead = mort[mort[1] == 1][[2, 10]].astype(int)
    dead.columns = ["MORTSTAT", "PERMTH_EXM"]
    # clumsy hack since 2012 doesn't have the CRP data. Will remove pending refactor of loading code
    if year == 2010:
        df = pd.concat([dem, gluc, cbc, crp, bioc, hdl, dead], axis=1).dropna()
    else:
        df = pd.concat([dem, gluc, cbc, bioc, hdl, dead], axis=1).dropna()
    df.index.name = "id"
    df = df.rename(
        {
            "RIDAGEYR": "age",
            "RIAGENDR": "sex",
            "LBDGLUSI": "glucose",
            "MORTSTAT": "is_dead",
            "PERMTH_EXM": "months_until_death",
            "LBXWBCSI": "white_blood_cell_count",
            "LBXLYPCT": "lymphocyte_percent",
            "LBXRDW": "red_blood_cell_distribution_width",
            "LBXMCVSI": "mean_cell_volume",
            "LBDLYMNO": "lymphocyte_number",
            "LBXRBCSI": "red_blood_cell_count",
            "LBXHGB": "hemoglobin",
            "LBXPLTSI": "platelet_count",
            "LBXMCHSI": "mean_cell_hemoglobin",
            "LBXBAPCT": "basophil_percent",
            "LBDHDDSI": "hdl_cholesterol",
            "LBXCRP": "c_reactive_protein",
            "LBDSALSI": "albumin",
            "LBDSCRSI": "creatinine",
            "LBXSAPSI": "alkaline_phosphate",
        },
        axis=1,
    )
    df = df.rename({"LB2RDW": "LBXRDW", "LB2WBCSI": "LBXWBCSI"}, axis=1)
    return df



import os
import shutil
import pytest
import time
from biolearn.cache import LocalFolderCache


@pytest.fixture
def cache():
    cache_path = "test_cache"
    max_size_gb = 0.000000954  # 1 KB for testing purposes
    cache = LocalFolderCache(cache_path, max_size_gb)
    yield cache
    # Clean up the cache directory after each test
    shutil.rmtree(cache_path, ignore_errors=True)


def test_store_and_get(cache):
    cache.store("key1", "value1")
    assert cache.get("key1") == "value1"


def test_get_nonexistent_key(cache):
    assert cache.get("nonexistent_key") is None


def test_store_and_get_multiple_keys(cache):
    cache.store("key1", "value1")
    cache.store("key2", "value2")
    assert cache.get("key1") == "value1"
    assert cache.get("key2") == "value2"


def test_cache_size_limit(cache):
    # Store items exceeding the cache size limit
    cache.store("key1", "a" * 500)  # 500 bytes
    time.sleep(0.1)  # Ensure file timestamps are different
    cache.store("key2", "b" * 800)  # 800 bytes
    time.sleep(0.1)
    cache.store("key3", "c" * 150)  # 150 bytes

    # Check that the least recently used item (key1) is evicted
    assert cache.get("key1") is None
    assert cache.get("key2") == "b" * 800
    assert cache.get("key3") == "c" * 150


def test_lru_respects_get_calls(cache):
    # Store items exceeding the cache size limit
    cache.store("key1", "a" * 500)
    time.sleep(0.1)  # Ensure file timestamps are different
    cache.store("key2", "b" * 400)
    time.sleep(0.1)
    cache.get("key1")  # Ensure Key1 most recently used
    time.sleep(0.1)
    cache.store("key3", "c" * 150)

    # Check that the least recently used item (key1) is evicted
    assert cache.get("key1") == "a" * 500
    assert cache.get("key2") is None
    assert cache.get("key3") == "c" * 150


def test_clear(cache):
    cache.store("key1", "value1")
    cache.store("key2", "value2")
    cache.clear()
    assert cache.get("key1") is None
    assert cache.get("key2") is None


def test_store_item_larger_than_cache_size(cache):
    # Store a small item in the cache
    small_item = "small"
    cache.store("small_key", small_item)
    time.sleep(0.1)  # Ensure file timestamps are different

    large_item = "x" * 2000  # 2000 bytes
    cache.store("large_key", large_item)

    assert cache.get("large_key") is None
    assert cache.get("small_key") == small_item



from biolearn.model import model_definitions
from biolearn.util import get_data_file
import yaml
import csv
import os

def generate_models_csv(models, output_file="generated/model_table.csv"):
    header = ["Name", "Year", "Species", "Tissue", "Predicts", "Source", "Coefficients"]

    # Generate the CSV data
    with open(output_file, "w", newline='') as csvfile:
        writer = csv.writer(csvfile)

        # Write the header
        writer.writerow(header)

        # Add each model to the CSV
        for name, details in models.items():
            if details["source"] == "unknown":
                source_link = "unknown"
            else:
                source_link = f"`paper <{details['source']}>`_"
            coefficients_link = f"`coefficients file <https://github.com/bio-learn/biolearn/blob/master/biolearn/data/{details['model']['file']}>`_"

            row = [
                name,
                details["year"],
                details["species"],
                details["tissue"],
                details.get('output', 'unknown'),
                source_link,
                coefficients_link
            ]
            writer.writerow(row)

    print(f"CSV generated at: {output_file}")

def ensure_folder_exists(path):
    # Get the directory part of the path (if it is a file)
    dir_path = os.path.dirname(path)

    if os.path.isdir(path):
        dir_path = path

    if not os.path.exists(dir_path):
        os.makedirs(dir_path)


def generate_data_csv_from_yaml(yaml_file, output_file="generated/data_table.csv"):
    header = ["ID", "Title", "Format", "Samples", "Age Present", "Sex Present"]

    with open(yaml_file, 'r') as stream:
        yaml_data = yaml.safe_load(stream)

    with open(output_file, "w", newline='') as csvfile:
        writer = csv.writer(csvfile)
        writer.writerow(header)

        for item in yaml_data["items"]:
            age_present = "Yes" if "age" in item.get("parser", {}).get("metadata", {}) else "No"
            sex_present = "Yes" if "sex" in item.get("parser", {}).get("metadata", {}) else "No"
            geo_link = f"`{item['id']} <https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc={item['id']}>`_"
            title = item['title'][:60] + '...' if len(item['title']) > 60 else item['title']
            row = [
                geo_link,
                title,
                item["format"],
                item["samples"],
                age_present,
                sex_present
            ]
            writer.writerow(row)

    print(f"CSV generated at: {output_file}")

if __name__ == "__main__":
    ensure_folder_exists("generated/")
    generate_models_csv(model_definitions)
    generate_data_csv_from_yaml(get_data_file("library.yaml"))



# biolearn documentation build configuration file, created by
# sphinx-quickstart on Fri Jan  8 09:13:42 2010.
#
# This file is execfile()d with the current directory set to its containing
# dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import os
import re
import sys

# import sphinx
# from biolearn._utils import _compare_version

# ----------------------------------------------------------------------------


# If extensions (or modules to document with autodoc) are in another
# directory, add these directories to sys.path here. If the directory
# is relative to the documentation root, use os.path.abspath to make it
# absolute, like shown here.
sys.path.insert(0, os.path.abspath("sphinxext"))
from github_link import make_linkcode_resolve

# We also add the directory just above to enable local imports of biolearn
sys.path.insert(0, os.path.abspath(".."))

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx_gallery.gen_gallery",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.imgmath",
    "sphinx.ext.intersphinx",
    'sphinx.ext.napoleon',
    "sphinxcontrib.bibtex",
    "sphinx.ext.linkcode",
    "gh_substitutions",
    "sphinx_copybutton",
    "sphinxext.opengraph",
    "myst_parser",
    "sphinx_design",
    "sphinx.ext.extlinks",
]

napoleon_google_docstring = True
autosummary_generate = True

autodoc_default_options = {
    "imported-members": True,
    "inherited-members": True,
    "undoc-members": True,
    "member-order": "bysource",
    #  We cannot have __init__: it causes duplicated entries
    #  'special-members': '__init__',
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["templates"]

# generate autosummary even if no references
autosummary_generate = True

# The suffix of source filenames.
source_suffix = [".rst", ".md"]

# The encoding of source files.
# source_encoding = 'utf-8'

# Generate the plots for the gallery
plot_gallery = "True"

# The master toctree document.
master_doc = "index"

# sphinxcontrib-bibtex
bibtex_bibfiles = ["./references.bib", "./soft_references.bib"]
bibtex_style = "unsrt"
bibtex_reference_style = "author_year"
bibtex_footbibliography_header = ""

# General information about the project.
project = "Biolearn"
copyright = "The biolearn developers 2023-2024"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.

# Latest release version
latest_release = '0.0.0'
# latest_release = re.match(
#     r"v?([0-9]+.[0-9]+.[0-9]+).*",
#     os.popen("git describe --tags").read().strip(),
# ).groups()[0]

# The full current version, including alpha/beta/rc tags.
import biolearn

current_version = biolearn.__version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
language = "en"

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of documents that shouldn't be included in the build.
# unused_docs = []
exclude_patterns = [
    "tune_toc.rst",
    "includes/big_toc_css.rst",
    "includes/bigger_toc_css.rst",
]

# List of directories, relative to source directory, that shouldn't be
# searched for source files.
exclude_trees = ["_build", "templates", "includes"]

# The reST default role (used for this markup: `text`) to use for all
# documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = False

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
# pygments_style = 'friendly'
# pygments_style = 'manni'
pygments_style = "sas"
pygments_dark_style = "stata-dark"


# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
html_theme = "furo"

# Add custom css instructions from themes/custom.css
html_css_files = [
    "custom.css",
    (
        "https://cdnjs.cloudflare.com/ajax/libs/"
        "font-awesome/5.15.4/css/all.min.css"
    ),
]

# The style sheet to use for HTML and HTML Help pages. A file of that name
# must exist either in Sphinx' static/ path, or in one of the custom paths
# given in html_static_path.
# html_style = 'nature.css'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    "dark_css_variables": {
        "color-announcement-background": "#935610",
        "color-announcement-text": "#FFFFFF",
    },
    "light_css_variables": {
        "admonition-font-size": "100%",
        "admonition-title-font-size": "100%",
        "color-announcement-background": "#FBB360",
        "color-announcement-text": "#111418",
        "color-admonition-title--note": "#448aff",
        "color-admonition-title-background--note": "#448aff10",
    },
}

# Add banner in case version is not stable
if "dev" in current_version:
    html_theme_options["announcement"] = (
        "<p>This is the development documentation "
        f"of biolearn ({current_version}) "
    )

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = ['themes']

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
html_title = "Biolearn"

# A shorter title for the navigation bar.  Default is the same as html_title.
html_short_title = "Biolearn"

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "logos/biolearn-transparent.png"

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "logos/favicon.ico"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["images", "themes"]

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
html_use_modindex = False

# If false, no index is generated.
html_use_index = False

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# variables to pass to HTML templating engine
build_dev_html = bool(int(os.environ.get("BUILD_DEV_HTML", False)))

html_context = {"build_dev_html": build_dev_html}

# If nonempty, this is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = ''

# Output file base name for HTML help builder.
htmlhelp_basename = "PythonScientic"

# Sphinx copybutton config
copybutton_prompt_text = ">>> "

# -- Options for LaTeX output ------------------------------------------------

# The paper size ('letter' or 'a4').
# latex_paper_size = 'letter'

# The font size ('10pt', '11pt' or '12pt').
# latex_font_size = '10pt'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto/manual]).
latex_documents = [
    (
        "index",
        "biolearn.tex",
        "Biomarkers with scikit-learn",
        "Mahdi, Martin and Seth"
        + r"\\\relax ~\\\relax http://biolearn.github.io",
        "manual",
    ),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
latex_logo = "logos/biolearn-transparent.png"

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# Additional stuff for the LaTeX preamble.

# Documents to append as an appendix to all manuals.
# latex_appendices = []
latex_elements = {
    "classoptions": ",oneside",
    "babel": "\\usepackage[english]{babel}",
    # Get completely rid of index
    "printindex": "",
}

# if _compare_version(sphinx.__version__, "<", "1.5"):
#     latex_preamble = r"""
#     \usepackage{amsmath}\usepackage{amsfonts}\usepackage{bm}\usepackage{morefloats}
#     \let\oldfootnote\footnote
#     \def\footnote#1{\oldfootnote{\small #1}}
#     """
#     # If false, no module index is generated.
#     latex_use_modindex = False
# else:
#     latex_elements[
#         "preamble"
#     ] = r"""
#     \usepackage{amsmath}\usepackage{amsfonts}\usepackage{bm}\usepackage{morefloats}
#     \let\oldfootnote\footnote
#     \def\footnote#1{\oldfootnote{\small #1}}
#     """


latex_domain_indices = False

# Show the page numbers in the references
latex_show_pagerefs = True

# Show URLs in footnotes
latex_show_urls = "footnote"

trim_doctests_flags = True

_python_doc_base = "https://docs.python.org/3.10"

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "python": (_python_doc_base, None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "scipy": ("https://scipy.github.io/devdocs/", None),
    "matplotlib": ("https://matplotlib.org/stable/", None),
    "sklearn": ("https://scikit-learn.org/stable/", None),
    "nibabel": ("https://nipy.org/nibabel", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable/", None),
    "nistats": ("https://nistats.github.io", None),
    "joblib": ("https://joblib.readthedocs.io/en/latest/", None),
}

extlinks = {
    "BoA": ("https://agingconsortium.org/%s", None),
    "sklearn": ("https://scikit-learn.org/stable/%s", None),
    "inria": ("https://team.inria.fr/%s", None),
    "biolearn-gh": ("https://github.com/biolearn/biolearn/%s", None),
    "neurostars": ("https://neurostars.org/tag/biolearn/%s", None),
    "nipy": ("https://nipy.org/%s", None),
}

binder_branch = "main" if "dev" in current_version else current_version

sphinx_gallery_conf = {
    "doc_module": "biolearn",
    "backreferences_dir": os.path.join("modules", "generated"),
    "reference_url": {"biolearn": None},
    "junit": "../test-results/sphinx-gallery/junit.xml",
    "examples_dirs": "../examples/",
    "gallery_dirs": "auto_examples",
    # Ignore the function signature leftover by joblib
    "ignore_pattern": r"func_code\.py",
    "show_memory": not sys.platform.startswith("win"),
    "remove_config_comments": True,
    "nested_sections": True,
    "binder": {
        "org": "bio-learn",
        "repo": "biolearn",
        "binderhub_url": "https://mybinder.org",
        "branch": binder_branch,
        "dependencies": "./binder/requirements.txt",
        "use_jupyter_lab": True,
    },
    "default_thumb_file": "logos/biolearn-transparent.png",
}


def touch_example_backreferences(app, what, name, obj, options, lines):
    # generate empty examples files, so that we don't get
    # inclusion errors if there are no examples for a class / module
    examples_path = os.path.join(
        app.srcdir, "modules", "generated", f"{name}.examples"
    )
    if not os.path.exists(examples_path):
        # touch file
        open(examples_path, "w").close()


def setup(app):
    app.connect("autodoc-process-docstring", touch_example_backreferences)


# The following is used by sphinx.ext.linkcode to provide links to github
linkcode_resolve = make_linkcode_resolve(
    "biolearn",
    "https://github.com/bio-learn/"
    "biolearn/blob/{revision}/"
    "{package}/{path}#L{lineno}",
)

# -- sphinxext.opengraph configuration -------------------------------------
ogp_site_url = "https://bio-learn.github.io/"
ogp_image = "https://bio-learn.github.io/_static/biolearn-logo.png"
ogp_use_first_image = True
ogp_site_name = "Biolearn"

autodoc_preserve_defaults=True


import pandas as pd
from biolearn.model import (
    model_definitions,
    LinearMethylationModel,
    GrimageModel,
    SexEstimationModel,
    ImputationDecorator,
    DeconvolutionModel,
)
from biolearn.imputation import (
    hybrid_impute,
    impute_from_average,
)
from biolearn.util import get_data_file


class ModelGallery:
    """
    ModelGallery manages a collection of Models that can be run on biological data to produce predictions.
    It supports retrieving models by name with various imputation methods, and searching models based on species or tissue.
    """

    model_builders = {
        "LinearMethylationModel": LinearMethylationModel.from_definition,
        "GrimageModel": GrimageModel.from_definition,
        "SexEstimationModel": SexEstimationModel.from_definition,
        "DeconvolutionModel": DeconvolutionModel.from_definition,
    }

    def __init__(self, models=model_definitions):
        """
        Initializes the ModelGallery instance.

        Args:
            models (dict): A dictionary of model definitions.
        """
        self.model_definitions = {}
        for name, model_def in models.items():
            if not isinstance(model_def, dict):
                raise ValueError(
                    f"Expected dictionary for model definition, got {type(model_def)} for model {name}"
                )
            model_type = model_def["model"]["type"]
            if model_type in self.model_builders:
                self.model_definitions[name] = model_def
            else:
                raise ValueError(
                    f"Model type {model_type} does not have a known builder"
                )

    def get(self, name, imputation_method=None):
        """
        Retrieves a model by its name with a specified imputation method to handle missing data.

        Args:
            name (str): The name of the model.
            imputation_method (str, optional): The method used for imputing missing data in the model's dataset.

        Returns:
            object: The requested model instance, possibly enhanced with an imputation strategy if specified.

        Raises:
            KeyError: If the model with the specified name is not found in the library.
            ValueError: If an invalid imputation method is specified.

        Available imputation_method selections are:
            - "none": No imputation is applied.
            - "averaging": Imputes missing values by calculating the average of data in the input set.
            - "dunedin": Uses hybrid_impute with the dunedinPACE gold standard values.
            - "sesame_450k": Uses hybrid_impute with the sesame 450k gold standard values.
        If no imputation method is selected then the default imputation method specified within the model's definition is used.
        """
        if name not in self.model_definitions:
            raise KeyError(f"Model not found: {name}")

        model_def = self.model_definitions[name]
        model_type = model_def["model"]["type"]
        model_instance = self.model_builders[model_type](model_def)

        global_default = "sesame_450k"
        default_imputation_method = model_def["model"].get(
            "default_imputation", global_default
        )
        imputation_method = imputation_method or default_imputation_method

        if imputation_method == "none":
            return model_instance
        elif imputation_method == "averaging":
            return ImputationDecorator(
                model_instance,
                lambda dnam, cpgs: impute_from_average(dnam, cpgs),
            )
        elif imputation_method == "dunedin":
            # DunedinPACE method
            gold_means_file = get_data_file("DunedinPACE_Gold_Means.csv")
            df = pd.read_csv(gold_means_file, index_col=0)
            gold_averages = df["mean"]
            return ImputationDecorator(
                model_instance,
                lambda dnam, cpgs: hybrid_impute(dnam, gold_averages, cpgs),
            )
        elif imputation_method == "sesame_450k":
            # Sesame method for 450k arrays
            sesame_450k_file = get_data_file("sesame_450k_median.csv")
            df = pd.read_csv(sesame_450k_file, index_col=0)
            gold_medians = df["median"]
            return ImputationDecorator(
                model_instance,
                lambda dnam, cpgs: hybrid_impute(dnam, gold_medians, cpgs),
            )
        else:
            raise ValueError(f"Invalid imputation method: {imputation_method}")

    def search(self, species=None, tissue=None):
        """
        Search models based on species and tissue criteria.

        Args:
            species (str, optional): The species to filter models by.
            tissue (str, optional): The tissue to filter models by.

        Returns:
            dict: A dictionary of models that match the specified criteria.
        """
        matches = {}
        for name, model_def in self.model_definitions.items():
            if (species is None or model_def.get("species") == species) and (
                tissue is None or model_def.get("tissue") == tissue
            ):
                matches[name] = model_def
        return matches



import appdirs
from biolearn.cache import LocalFolderCache


def default_cache():
    app_name = "bio-learn"
    cache_path = appdirs.user_cache_dir(app_name)
    return LocalFolderCache(cache_path, 50)



import yaml
import pandas as pd
import numpy as np
import re
from biolearn.util import cached_download, get_data_file
from biolearn.defaults import default_cache
from biolearn.cache import NoCache


def parse_after_colon(s):
    """Extract and return the substring after the first colon."""

    if isinstance(s, str) and s.strip() and ":" in s:
        return s.split(":")[1].strip()

    return ""


def sex_parser(s):
    if isinstance(s, str):
        s_lower = s.lower().strip()
        if s_lower in ["female", "f"]:
            return 1
        elif s_lower in ["male", "m"]:
            return 2
    return 0


def extract_numeric(s):
    """Extract the first numeric value from a string."""
    match = re.search(r"(\d+(\.\d+)?)", s)
    return float(match.group(1)) if match else None


class QualityReport:
    def __init__(self, sample_report, methylation_site_report, summary):
        self.sample_report = sample_report
        self.methylation_site_report = methylation_site_report
        self.summary = summary

    def show(self):
        """
        Pretty prints the summary of the quality report for display, along with relevant notes based on the data.
        """
        print("Quality Report Summary\n")
        print("------------------------------------------------")

        total_samples = self.summary["total_samples"]
        total_sites = self.summary["methylation_sites"]
        missing_data_points = self.summary["missing_data_points"]

        total_data_points = total_samples * total_sites
        missing_data_percentage = (
            (missing_data_points / total_data_points) * 100
            if total_data_points > 0
            else 0
        )
        print(f"Sample Count: {total_samples}")
        print(f"Methylation Sites: {total_sites}")
        print(
            f"Missing Methylation Data: {missing_data_points} ({missing_data_percentage:.2f}%)"
        )

        high_deviation_count = self.summary["samples_with_high_deviation"]
        high_deviation_percentage = (
            high_deviation_count / total_samples
        ) * 100
        print(
            f"Samples With High Deviation: {high_deviation_count} ({high_deviation_percentage:.2f}%)"
        )

        sites_with_missing_data = self.summary[
            "sites_with_over_20_percent_missing"
        ]
        missing_data_sites_percentage = (
            sites_with_missing_data / total_sites
        ) * 100
        print(
            f"Methylation Sites With Over 20% of Reads Missing: {sites_with_missing_data} ({missing_data_sites_percentage:.2f}%)"
        )

        print("\nNotes:")
        print("------------------------------------------------")
        if missing_data_points == 0:
            print(
                "- No missing data points implies that this data has already gone through an imputation process or that low quality reads were included."
            )

        if sites_with_missing_data > 0:
            print(
                "- Your data set includes methylation sites that have over 20% of reads missing. Default imputation may replace the values for all reads from this site with a gold standard."
            )

        if high_deviation_count > 0:
            print(
                "- Your data set includes samples with a high deviation. It is likely that the methylation data for these samples has been distorted due to technical issues."
            )
        print("\n")


class GeoData:
    """
    Represents genomic data with a focus on metadata and methylation data.

    GeoData facilitates the organization and access to metadata and methylation data.

    Attributes:
        metadata (DataFrame): A pandas DataFrame where rows represent different samples
                              and columns represent different data fields.
        dnam (DataFrame): A pandas DataFrame where columns represent different samples
                          and rows represent different methylation sites.
    """

    def __init__(self, metadata, dnam):
        """
        Initializes the GeoData instance.

        Args:
            metadata (DataFrame): Metadata associated with genomic samples.
            dnam (DataFrame): Methylation data associated with genomic samples.
        """
        self.metadata = metadata
        self.dnam = dnam

    def copy(self):
        """
        Creates a deep copy of the GeoData instance.

        Returns:
            GeoData: A new instance of GeoData with copies of the metadata and dnam DataFrames.
        """
        return GeoData(
            self.metadata.copy(deep=True), self.dnam.copy(deep=True)
        )

    def quality_report(self, sites=None):
        """
        Generates a quality control report for the genomic data, optionally filtered by specified methylation sites,
        and includes a detailed section reporting the missing percentage for each methylation site.

        Args:
            sites (list, optional): A list of methylation site identifiers to include in the report.
                                    If None, all sites are included.

        Returns:
            QualityReport: An object containing both detailed methylation data, a summary,
                           and a detailed section for missing percentages per site.
        """
        methylation_data = self.dnam.copy()

        # Filter methylation data if specific sites are provided
        if sites is not None:
            methylation_data = methylation_data.loc[sites]

        # Calculate mean absolute deviation
        methylation_medians = methylation_data.median(axis=1)
        deviations = methylation_data.sub(methylation_medians, axis=0)
        mean_abs_deviation = deviations.abs().mean()
        sample_report = mean_abs_deviation.to_frame(name="deviation")

        # Include 'age' in the detailed report if relevant and if analyzing all sites
        if "age" in self.metadata.columns and sites is None:
            sample_report["age"] = self.metadata["age"]

        # Calculate missing percentage for each methylation site
        missing_percentage = methylation_data.isna().mean(axis=1)
        methylation_site_report = missing_percentage.to_frame(name="missing")

        # Summary calculations
        total_nan = methylation_data.isna().sum().sum()
        sites_with_20_percent_nan = (
            methylation_data.isna().mean(axis=1) >= 0.2
        ).sum()

        summary = {
            "missing_data_points": total_nan,
            "sites_with_over_20_percent_missing": sites_with_20_percent_nan,
            "samples_with_high_deviation": (mean_abs_deviation > 0.04).sum(),
            "total_samples": methylation_data.shape[1],
            "methylation_sites": methylation_data.shape[0],
        }

        return QualityReport(sample_report, methylation_site_report, summary)

    @classmethod
    def from_methylation_matrix(cls, matrix):
        """
        Creates a GeoData instance from a methylation matrix which can be
        either a DataFrame directly or a path to a CSV file.

        Args:
            matrix (Union[str, DataFrame]): Methylation matrix as a DataFrame or the path to the CSV file containing the matrix.

        Returns:
            GeoData: An instance of GeoData with the methylation data loaded and metadata initialized.
        """
        if isinstance(matrix, str):
            # If the input is a string, assume it's a file path and read the CSV
            dnam = pd.read_csv(matrix, index_col=0)
        elif isinstance(matrix, pd.DataFrame):
            # If the input is already a DataFrame, use it directly
            dnam = matrix
        else:
            raise ValueError(
                "The matrix must be either a DataFrame or a file path to a CSV."
            )

        # Process row identifiers
        dnam.index = dnam.index.str.split("_").str[0]

        # Combine duplicate rows by averaging their values
        dnam = dnam.groupby(dnam.index).mean()

        # Create an empty DataFrame for metadata with row identifiers as columns
        metadata = pd.DataFrame(index=dnam.columns)

        return cls(metadata, dnam)


class GeoMatrixParser:
    parsers = {
        "numeric": lambda s: extract_numeric(parse_after_colon(s)),
        "string": lambda s: parse_after_colon(s),
        "sex": lambda s: sex_parser(parse_after_colon(s)),
    }
    seperators = {"space": " ", "comma": ","}

    def __init__(self, data):
        if data.get("id-row") is None:
            raise ValueError("Parser not valid: missing id-row")
        self.id_row = data.get("id-row")
        self.metadata = data.get("metadata")
        self.matrix_start = data.get("matrix-start")
        self.matrix_file = data.get("matrix-file")
        self.matrix_file_seperator = self.seperators.get(
            data.get("matrix-file-seperator")
        )
        self.matrix_file_key_line = data.get("matrix-file-key-line")

    def parse(self, file_path):
        load_list = self._metadata_load_list()
        load_rows = [x[1] for x in load_list]
        column_names = [x[0] for x in load_list]
        metadata = pd.read_table(
            file_path,
            index_col=0,
            skiprows=lambda x: x != self.id_row - 1 and x not in load_rows,
        )
        metadata.index = column_names
        metadata = metadata.transpose()
        metadata.index.name = "id"
        for col in metadata.columns:
            parser_name = self.metadata[col]["parse"]
            parser = self.parsers[parser_name]
            metadata[col] = metadata[col].apply(parser)
        if self.matrix_start:
            dnam = pd.read_table(
                file_path, index_col=0, skiprows=self.matrix_start - 1
            )
            dnam = dnam.drop(["!series_matrix_table_end"], axis=0)
            dnam.index.name = "id"
        elif self.matrix_file:
            matrix_file_path = cached_download(self.matrix_file)
            print(
                f"Note: This dataset will take a few minutes to load and may fail if you have insufficient memory"
            )
            df = pd.read_csv(
                matrix_file_path, index_col=0, sep=self.matrix_file_seperator
            )
            methylation_df = df.iloc[:, ::2]
            pval_df = df.iloc[:, 1::2]
            pval_df = pval_df.replace("<1E-16", "0", regex=False).astype(
                float, errors="ignore"
            )
            pval_df = pval_df.map(lambda x: np.nan if x > 0.05 else 0)
            # NaN values in pval_df will cause corresponding values in methylation_df to be NaN
            dnam = methylation_df + pval_df.values
            dnam = self._remap_and_prune_columns(dnam, file_path)
        return GeoData(metadata, dnam)

    def _remap_and_prune_columns(self, data, matrix_file_path):
        mapping_df = pd.read_table(
            matrix_file_path,
            index_col=0,
            skiprows=lambda x: x != self.id_row - 1
            and x != self.matrix_file_key_line - 1,
        )
        column_mapping = mapping_df.to_dict("records")[0]

        # Reverse the mapping if needed as key is based on first line loaded
        reverse_mapping = self.id_row < self.matrix_file_key_line
        if reverse_mapping:
            column_mapping = {v: k for k, v in column_mapping.items()}

        data = data.rename(columns=column_mapping)
        data = data[
            [col for col in data.columns if col in column_mapping.values()]
        ]
        return data

    def _metadata_load_list(self):
        load_list = [
            (key, self.metadata[key]["row"] - 1)
            for key in self.metadata.keys()
        ]
        load_list.sort(key=lambda x: x[1])
        return load_list


class DataSource:
    """
    Represents a single data source in the DataLibrary.

    This class encapsulates the details of a data source including metadata about the source and
    functionality to load the data.

    Raises:
        ValueError: If any of the required fields ('id', 'path', 'parser') are missing
                    during initialization.
    """

    REQUIRED_FIELDS = {
        "id": "'id' key is missing in item",
        "path": "'path' key is missing in item",
        "parser": "'parser' key is missing in item",
    }

    def __init__(self, data, cache=None):
        """
        Initializes the DataSource instance with configuration data and an optional cache mechanism.
        This method parses a dictionary typically loaded from a YAML configuration file for a data source.
        It checks for essential fields, sets up attributes, and configures a parser for data handling.

        Args:
            data (dict): A dictionary containing the data source's properties. Must include keys like 'id',
                        'path', 'parser', and optionally 'title', 'summary', 'format', and 'organism'.
            cache (object, optional): An object that adheres to the caching interface used in the caching module.
                        If no cache is provided, a default NoCache instance is used.

        Raises:
            ValueError: If any of the required fields ('id', 'path', 'parser') are missing in the input data.
        """
        for field, error_message in self.REQUIRED_FIELDS.items():
            setattr(self, field, data.get(field))
            if getattr(self, field) is None:
                raise ValueError(error_message)
        self.cache = cache if cache else NoCache()
        self.title = data.get(
            "title", ""
        )  # Default empty string if title is not provided
        self.summary = data.get(
            "summary", ""
        )  # Default empty string if summary is not provided
        self.format = data.get(
            "format", ""
        )  # Default empty string if format is not provided
        self.organism = data.get(
            "organism", ""
        )  # Default empty string if organism is not provided

        self.parser = self._create_parser(data["parser"])

    def load(self):
        """
        Loads the data from the source.

        Returns:
            GeoData: An instance of the GeoData class containing the parsed geographical data.
        """
        cached = self.cache.get(self.id)
        if cached:
            return cached
        else:
            data = self.parser.parse(self.path)
            self.cache.store(self.id, data)
            return data

    def __repr__(self):
        """
        Represents the DataSource instance as a string.

        Returns:
            str: A string representation of the DataSource instance.
        """
        return f"DataSource(ID: {self.id}, Title: {self.title})"

    def _create_parser(self, parser_data):
        parser_type = parser_data.get("type")
        if parser_type is None:
            raise ValueError("Parser type is missing")
        if parser_type == "geo-matrix":
            return GeoMatrixParser(parser_data)
        # Add more parsers as needed
        raise ValueError(f"Unknown parser type: {parser_type}")


def parse_library_file(library_file, cache=None):
    data = yaml.safe_load(library_file)
    if data is None:
        raise ValueError("File must be YAML format with 'items' at root")
    if "items" in data:
        data_sources = [
            DataSource(item, cache if cache else NoCache())
            for item in data["items"]
        ]
        return data_sources
    else:
        raise ValueError("File must be YAML format with 'items' at root")


class DataLibrary:
    """
    Manages a collection of data sources for biomarkers research.

    The DataLibrary class is responsible for loading, storing, and retrieving data sources.
    Data sources are defined in a library file and new sources can easily be added at runtime.
    Currently DNA methylation data from GEO is supported.
    """

    def __init__(self, library_file=None, cache=None):
        """
        Initializes the DataLibrary instance with an optional library file and cache mechanism.

        Args:
            library_file (str, optional): The path to the library file. If None, the default biolearn
                                          library file is loaded.
            cache (object, optional): An object that adheres to the caching interface used in the caching module. If None, the
                                      default cache is used. This cache will be used by all returned
                                      data sources
        """
        self.cache = cache if cache else default_cache()
        self.sources = []
        if library_file is None:
            library_file = get_data_file("library.yaml")
        self.load_sources(library_file)

    def load_sources(self, library_file):
        """
        Loads data sources from a given library file appending them to the current set of data sources.

        Args:
            library_file (str): The file path of the library file to load data sources from.
        """
        with open(library_file, "r") as f:
            data_sources = self._parse_library_file(f)
            self.sources.extend(data_sources)

    def get(self, source_id):
        """
        Retrieves a data source by its identifier.

        Args:
            source_id (str): The identifier of the data source to retrieve.

        Returns:
            The data source with the given identifier if found, otherwise None.
        """
        for source in self.sources:
            if source.id == source_id:
                return source
        return None

    def lookup_sources(self, organism=None, format=None):
        """
        Looks up data sources based on the specified organism and/or format.

        Args:
            organism (str, optional): The organism to filter the data sources by.
            format (str, optional): The format to filter the data sources by.

        Returns:
            A list of data sources that match the specified organism and format criteria.
        """
        matches = []
        for source in self.sources:
            if (organism is None or source.organism == organism) and (
                format is None or source.format == format
            ):
                matches.append(source)
        return matches

    def _parse_library_file(self, library_file):
        return parse_library_file(library_file, self.cache)



import pytest
import os
import pandas as pd
import numpy as np
from io import StringIO
from biolearn.data_library import (
    parse_library_file,
    DataSource,
    DataLibrary,
    GeoData,
)
from biolearn.model import get_data_file
from biolearn.util import get_test_data_file
from biolearn.util import load_test_data_file, get_test_data_file

import pandas as pd


def test_quality_report():
    sample_inputs = load_test_data_file("external/DNAmTestSet.csv")
    sample_metadata = load_test_data_file("external/testset_metadata.csv")
    geo_data_instance = GeoData(sample_metadata, sample_inputs)
    actual_report = geo_data_instance.quality_report()

    expected_report = load_test_data_file("test_quality_report.csv")

    # Directly compare the actual and expected reports
    pd.testing.assert_frame_equal(
        actual_report.sample_report,
        expected_report,
        check_dtype=True,
        check_like=True,
    )


def test_blank_file_gives_error():
    file_contents = " "
    test_file = StringIO()
    test_file.write(file_contents)
    with pytest.raises(ValueError) as e:
        parse_library_file(test_file)
    assert str(e.value) == "File must be YAML format with 'items' at root"


def test_can_load_single_item_correctly():
    file_contents = """
---
items:
- id: GSE40279
  title: An informative title
  summary: A Useful Summary
  format: Illumina450k
  organism: human
  path: https://ftp.ncbi.nlm.nih.gov/geo/series/GSE40nnn/GSE40279/matrix/GSE40279_series_matrix.txt.gz
  parser:
    type: geo-matrix
    id-row: 12
    metadata:
      age:
        row: 44
        parse: numeric
    matrix-start: 72
"""
    test_file = StringIO(file_contents)
    actual = parse_library_file(test_file)

    assert len(actual) == 1
    item = actual[0]
    assert item.id == "GSE40279"
    assert item.title == "An informative title"
    assert item.format == "Illumina450k"
    assert item.organism == "human"
    assert (
        item.path
        == "https://ftp.ncbi.nlm.nih.gov/geo/series/GSE40nnn/GSE40279/matrix/GSE40279_series_matrix.txt.gz"
    )
    parser = item.parser
    assert parser.id_row == 12
    assert parser.metadata == {"age": {"row": 44, "parse": "numeric"}}
    assert parser.matrix_start == 72


def test_missing_id_gives_error():
    file_contents = """
---
items:
- title: An informative title
  path: https://ftp.ncbi.nlm.nih.gov/geo/series/GSE40nnn/GSE40279/matrix/GSE40279_series_matrix.txt.gz
  parser:
    type: geo-matrix
    id-row: 12
    metadata:
      age:
        row: 44
        parse: numeric
    matrix-start: 72
"""
    test_file = StringIO(file_contents)
    with pytest.raises(ValueError) as e:
        parse_library_file(test_file)
    assert str(e.value) == "'id' key is missing in item"


def test_unknown_parser_type_gives_error():
    file_contents = """
---
items:
- id: GSE40279
  path: https://ftp.ncbi.nlm.nih.gov/geo/series/GSE40nnn/GSE40279/matrix/GSE40279_series_matrix.txt.gz
  parser:
    type: unknown-parser
    id-row: 12
    metadata:
      age:
        row: 44
        parse: numeric
    matrix-start: 72
"""
    test_file = StringIO(file_contents)
    with pytest.raises(ValueError) as e:
        parse_library_file(test_file)
    assert str(e.value) == "Unknown parser type: unknown-parser"


def test_missing_path_gives_error():
    file_contents = """
---
items:
- id: GSE40279
  parser:
    type: geo-matrix
    id-row: 12
    metadata:
      age:
        row: 44
        parse: numeric
    matrix-start: 72
"""
    test_file = StringIO(file_contents)
    with pytest.raises(ValueError) as e:
        parse_library_file(test_file)
    assert str(e.value) == "'path' key is missing in item"


def test_parser_missing_id_row_gives_error():
    file_contents = """
---
items:
- id: GSE40279
  path: https://ftp.ncbi.nlm.nih.gov/geo/series/GSE40nnn/GSE40279/matrix/GSE40279_series_matrix.txt.gz
  parser:
    type: geo-matrix
    misspelled-id-row: 12
    metadata:
      age:
        row: 44
        parse: numeric
    matrix-start: 72
"""
    test_file = StringIO(file_contents)
    with pytest.raises(ValueError) as e:
        parse_library_file(test_file)
    assert str(e.value) == "Parser not valid: missing id-row"


def test_missing_parser_gives_error():
    file_contents = """
---
items:
- id: GSE40279
  path: https://ftp.ncbi.nlm.nih.gov/geo/series/GSE40nnn/GSE40279/matrix/GSE40279_series_matrix.txt.gz
"""
    test_file = StringIO(file_contents)
    with pytest.raises(ValueError) as e:
        parse_library_file(test_file)
    assert str(e.value) == "'parser' key is missing in item"


def test_can_load_library_file():
    library_file = get_data_file("library.yaml")
    with open(library_file, "r") as the_file:
        actual = parse_library_file(the_file)

    assert len(actual) > 1


def test_can_load_dnam():
    data_source_spec = {
        "id": "TestData",
        "path": get_test_data_file("geo_dnam_test_file"),
        "parser": {
            "type": "geo-matrix",
            "id-row": 33,
            "metadata": {
                "age": {"row": 47, "parse": "numeric"},
                "sex": {"row": 41, "parse": "sex"},
                "cancer": {"row": 50, "parse": "string"},
            },
            "matrix-start": 74,
        },
    }
    source = DataSource(data_source_spec)

    df = source.load()
    # Verify data set is of known size
    dnam = df.dnam
    assert dnam.shape == (37, 5)
    assert df.metadata.shape == (5, 3)
    assert "cancer" in df.metadata.columns.to_list()
    assert np.issubdtype(df.metadata["age"], np.number)
    assert np.issubdtype(df.metadata["sex"], np.number)
    assert (df.metadata["sex"] != 0).all()
    assert all(
        np.issubdtype(dnam[col].dtype, np.number) for col in dnam.columns
    )


def test_load_dnam_with_matrix_file():
    data_source_spec = {
        "id": "TestMatrixFile",
        "path": get_test_data_file("geo_dnam_test_file"),
        "parser": {
            "type": "geo-matrix",
            "id-row": 33,
            "metadata": {
                "age": {"row": 47, "parse": "numeric"},
                "sex": {"row": 41, "parse": "sex"},
                "cancer": {"row": 50, "parse": "string"},
            },
            "matrix-file": get_test_data_file("test_supplementary_matrix.txt"),
            "matrix-file-seperator": "space",
            "matrix-file-key-line": 72,
        },
    }
    source = DataSource(data_source_spec)

    df = source.load()
    dnam = df.dnam
    metadata = df.metadata

    assert dnam.shape == (37, 5)
    assert metadata.shape == (5, 3)
    assert "cancer" in metadata.columns.to_list()
    assert np.issubdtype(metadata["age"], np.number)
    assert np.issubdtype(metadata["sex"], np.number)
    assert (metadata["sex"] != 0).all()
    assert all(
        np.issubdtype(dnam[col].dtype, np.number) for col in dnam.columns
    )
    assert dnam.isna().sum().sum() == 8
    assert all(col in metadata.index for col in dnam.columns)


def test_load_sources_append():
    # Initialize DataLibrary
    library = DataLibrary(
        library_file=get_test_data_file("library_files/library.yaml")
    )

    # Check if sources were loaded
    assert len(library.sources) == 2


def test_get_source_by_id():
    # Initialize DataLibrary
    library = DataLibrary(
        library_file=get_test_data_file("library_files/library.yaml")
    )

    # Get source by ID
    source = library.get("GSE40279")
    assert source.id == "GSE40279"
    assert (
        source.title
        == "Genome-wide Methylation Profiles Reveal Quantitative Views of Human Aging Rates"
    )


def test_lookup_sources_by_format():
    # Initialize DataLibrary
    library = DataLibrary(
        library_file=get_test_data_file("library_files/library.yaml")
    )

    matches = library.lookup_sources(organism="human", format="Illumina27k")
    assert len(matches) == 1
    assert matches[0].id == "GSE19711"


def test_lookup_sources_by_organism():
    # Initialize DataLibrary
    library = DataLibrary(
        library_file=get_test_data_file("library_files/library.yaml")
    )

    matches = library.lookup_sources(organism="human")
    assert len(matches) == 2



from docutils.nodes import reference
from docutils.parsers.rst.roles import set_classes

# adapted from
# https://github.com/mne-tools/mne-python/blob/main/doc/sphinxext/gh_substitutions.py


def _gh_role(name, rawtext, text, lineno, inliner, options={}, content=[]):
    """Link to a GitHub issue."""
    try:
        # issue/PR mode (issues/PR-num will redirect to pull/PR-num)
        int(text)
    except ValueError:
        # direct link mode
        slug = text
    else:
        slug = f"issues/{text}"
    text = f"#{text}"
    ref = f"https://github.com/bio-learn/biolearn/{slug}"
    set_classes(options)
    node = reference(rawtext, text, refuri=ref, **options)
    return [node], []


def setup(app):
    app.add_role("gh", _gh_role)



"""
"Phenotypic Ages" in NHANES Data
=============================================================

This example loads blood exam data from NHANES 2010, calculates "Phenotypic Ages," and performs survival analyses by phenotypic age. 
"""

#############################################################################
# Loading NHANES 2010 data
# ---------------------------------------
from biolearn.load import load_nhanes 
year = 2010
df = load_nhanes(year)
df["years_until_death"] = df["months_until_death"] / 12

#############################################################################
# Calculate "biological age" based on PhenotypicAge
# ------------------------------------------------------
from biolearn.hematology import phenotypic_age
df["phenotypic_age"]=phenotypic_age(df)

#############################################################################
# Show relation between biological age and chronological age
# ---------------------------------------------------------------
import seaborn as sn
sn.scatterplot(data=df,x="age", y="phenotypic_age",s=2);

############################################################################################################################
# Plot survival curve for people with accelerated aging (older biological age) vs decelerated aging (younger biological age)
# --------------------------------------------------------------------------------------------------------------------------
import matplotlib.pyplot as plt
from lifelines import KaplanMeierFitter
df["biologically_older"] = df["phenotypic_age"] > df["age"]
ax = plt.subplot()
groups = df["biologically_older"]
ix = groups == 0
T = df.years_until_death
E = df.is_dead
kmf = KaplanMeierFitter()
kmf.fit(T[ix], E[ix], label="Biologically younger")
ax = kmf.plot_survival_function(ax=ax)
kmf.fit(T[~ix], E[~ix], label="Biologically older")
ax = kmf.plot_survival_function()
plt.ylabel("Survival")
plt.xlabel("Years");


import os
import shutil
import pickle


class NoCache:
    """
    A cache implementation that does nothing.
    """

    def get(self, key):
        """
        Retrieves an item from the cache.

        Args:
            key (str): The key for which the cache value is retrieved.

        Returns:
            None: Always returns None, simulating a cache miss.
        """
        return None

    def store(self, key, value):
        """
        Stores an item in the cache.

        Args:
            key (str): The key under which the value is stored.
            value (any): The value to be stored.
        """
        pass

    def clear(self):
        """
        Clears all items from the cache.
        """
        pass


class LocalFolderCache:
    """
    A cache that stores data as serialized objects in a local directory. Implements basic LRU (Least Recently Used)
    cleanup to maintain the cache within a specified maximum size limit.
    """

    def __init__(self, path, max_size_gb):
        """
        Initializes the LocalFolderCache instance.

        Args:
            path (str): The filesystem path to the directory to be used for cache storage.
            max_size_gb (float): The maximum size of the cache in gigabytes.
        """
        self.path = path
        self.max_size_bytes = (
            max_size_gb * 1024 * 1024 * 1024
        )  # Convert GB to bytes
        os.makedirs(self.path, exist_ok=True)

    def get(self, key):
        """
        Retrieves an item from the cache by its key, if it exists.

        Args:
            key (str): The key for which the cache value is retrieved.

        Returns:
            any: The value stored in the cache, or None if the key does not exist or an error occurs.
        """
        file_path = os.path.join(self.path, key)
        if os.path.exists(file_path):
            try:
                with open(file_path, "rb") as file:
                    value = pickle.load(file)
                return value
            except (
                pickle.UnpicklingError,
                EOFError,
                FileNotFoundError,
                IOError,
            ):
                return None
        return None

    def store(self, key, value):
        """
        Stores an item in the cache under the specified key, managing the cache size to stay within limits.

        Args:
            key (str): The key under which the value is stored.
            value (any): The value to be stored.
        """
        file_path = os.path.join(self.path, key)
        serialized_value = pickle.dumps(value)
        value_size = len(serialized_value)
        if value_size > self.max_size_bytes:
            return
        with open(file_path, "wb") as file:
            file.write(serialized_value)
        self._cleanup()

    def _cleanup(self):
        """
        Cleans up the cache directory to maintain it within the maximum size limit by removing the least
        recently accessed files first. Called automatically everytime a new item is stored.
        """
        total_size = sum(
            os.path.getsize(os.path.join(self.path, f))
            for f in os.listdir(self.path)
        )
        if total_size > self.max_size_bytes:
            files = sorted(
                os.listdir(self.path),
                key=lambda f: os.path.getatime(os.path.join(self.path, f)),
                reverse=True,
            )
            while total_size > self.max_size_bytes and files:
                file_to_remove = files.pop()
                file_path = os.path.join(self.path, file_to_remove)
                total_size -= os.path.getsize(file_path)
                os.remove(file_path)

    def clear(self):
        """
        Clears the cache by removing the cache directory and all its contents.
        """
        shutil.rmtree(self.path, ignore_errors=True)
